"""Items to test functionality around the containers interface"""

from __future__ import annotations

from pathlib import Path

import pytest

from flint.containers import (
    KNOWN_CONTAINER_LOOKUP,
    LIST_OF_KNOWN_CONTAINERS,
    FlintContainer,
    _sanity_check_containers,
    download_known_containers,
    get_known_container_path,
    log_known_containers,
    verify_known_containers,
)

from .test_helpers import which


def test_sanity_check_containers():
    """Make sure that the reference containers are valid with a simple set of checks.
    Make sure not doubling up on the reference name of file name"""
    _sanity_check_containers(container_list=LIST_OF_KNOWN_CONTAINERS)
    from copy import deepcopy

    dummy = list(deepcopy(LIST_OF_KNOWN_CONTAINERS))
    dummy.append(dummy[0])

    with pytest.raises(AssertionError):
        _sanity_check_containers(container_list=dummy)


def test_verify_known_containers(tmpdir):
    """Check that our verify function works. This will cheat and
    crate some temporary file with the expected file name as the
     verify function in of itself (currently) only checks to see
     if a file exists"""
    container_directory = Path(tmpdir) / "containers1"
    container_directory.mkdir(parents=True)

    assert not verify_known_containers(container_directory=container_directory)

    for cata in LIST_OF_KNOWN_CONTAINERS:
        cata_path = container_directory / cata.file_name
        cata_path.touch()

    assert verify_known_containers(container_directory=container_directory)


def test_all_flint_containers():
    """Make sure everything we know is a FlintContainer"""
    assert all([isinstance(fc, FlintContainer) for fc in LIST_OF_KNOWN_CONTAINERS])


def test_all_known_containers():
    """Same as above but for the known donctainers lookup"""
    for k, v in KNOWN_CONTAINER_LOOKUP.items():
        assert isinstance(k, str)
        assert isinstance(v, FlintContainer)


def test_log_containers():
    """Output all the known containers"""
    # This should simply not error
    log_known_containers()


if which("singularity") is None:
    pytest.skip("Singularity is not installed", allow_module_level=True)


@pytest.fixture(scope="session")
def flint_containers(tmp_path_factory) -> Path:
    """Download all of the flint containers"""
    flint_container_path = Path(tmp_path_factory.mktemp("download_containers"))
    flint_container_path.mkdir(parents=True, exist_ok=True)

    container_paths = download_known_containers(
        container_directory=flint_container_path, new_tag=None
    )

    assert all(isinstance(path, Path) for path in container_paths)
    assert all(path.exists() for path in container_paths)

    return flint_container_path


def test_download_flint_containers(flint_containers) -> None:
    """Start the download of the flint containers"""
    assert isinstance(flint_containers, Path)
    assert flint_containers.exists()


def test_verify_containers_with_containers(flint_containers):
    """Make sure the actual containers downloaded register as correct. This
    uses the download fixture"""
    assert verify_known_containers(container_directory=flint_containers)


def test_get_known_container_path(flint_containers):
    """Get the to a known container"""
    casa_container = get_known_container_path(
        container_directory=flint_containers, name="casa"
    )
    assert isinstance(casa_container, Path)
    assert casa_container.exists()

    with pytest.raises(ValueError):
        casa_container = get_known_container_path(
            container_directory=flint_containers, name="JackSparrowNotBeKnown"
        )



"""Some tests around the sclient utility"""

from __future__ import annotations

from pathlib import Path

import pytest

from flint.sclient import pull_container, run_singularity_command

from .test_helpers import which

if which("singularity") is None:
    pytest.skip("Singularity is not installed", allow_module_level=True)


@pytest.fixture(scope="session")
@pytest.mark.flaky(reruns=3, reruns_delay=2)
def hello_world_container(tmp_path_factory) -> Path:
    """Download the hello world container once and use it across the session"""
    temp_container_dir = Path(tmp_path_factory.mktemp("hello_world"))
    temp_container_dir.mkdir(parents=True, exist_ok=True)

    name = "example.sif"
    temp_container = temp_container_dir / name

    assert not temp_container.exists()

    output_path = pull_container(
        container_directory=temp_container_dir,
        uri="docker://alpine:latest",
        file_name=name,
    )
    return output_path


def test_pull_apptainer(hello_world_container):
    """Attempt to pull down an example container"""

    assert hello_world_container.exists()
    assert isinstance(hello_world_container, Path)


def test_run_singularity_command(hello_world_container):
    """Make sure that the running of a container works"""
    run_singularity_command(image=hello_world_container, command="echo 'JackSparrow'")


def test_raise_error_no_container_noexist() -> None:
    """Should an incorrect path be given an error should be raised"""

    no_exists_container = Path("JackBeNotHereMate.sif")
    assert not no_exists_container.exists()

    with pytest.raises(FileNotFoundError):
        run_singularity_command(image=no_exists_container, command="PiratesbeHere")


def test_positive_max_retries() -> None:
    """Make sure an error is raised if `max_retries` reaches break case"""
    no_exists_container = Path("JackBeNotHereMate.sif")
    assert not no_exists_container.exists()

    # These errors are fired off before the check for the container is made
    with pytest.raises(ValueError):
        run_singularity_command(
            image=no_exists_container, command="PiratesbeHere", max_retries=0
        )
    with pytest.raises(ValueError):
        run_singularity_command(
            image=no_exists_container, command="PiratesbeHere", max_retries=-222
        )

    with pytest.raises(FileNotFoundError):
        run_singularity_command(
            image=no_exists_container, command="PiratesbeHere", max_retries=111
        )



"""Tooling around the use of the `addmodel` program packaged with `calibrate`.

This tooling requires the `calibrate` container.
"""

from __future__ import annotations

from argparse import ArgumentParser
from pathlib import Path
from typing import Literal

from flint.logging import logger
from flint.ms import remove_columns_from_ms
from flint.options import BaseOptions, add_options_to_parser, create_options_from_parser
from flint.sclient import run_singularity_command


class AddModelOptions(BaseOptions):
    """Container for options into the ``addmodel`` program packaged
    with ``aocalibrate``"""

    model_path: Path
    """Path to the sky-model that will be inserted"""
    ms_path: Path
    """Path to the measurement set that will be interacted with"""
    mode: Literal["a", "s", "c", "v"]
    """The mode ``addmodel`` will be operating under, where where a=add model to visibilities (default), s=subtract model from visibilities, c=copy model to visibilities, z=zero visibilities"""
    datacolumn: str
    """The column that will be operated against"""


def add_model_options_to_command(add_model_options: AddModelOptions) -> str:
    """Generate the command to execute ``addmodel``

    Args:
        add_model_options (AddModelOptions): The collection of supported options used to generate the command

    Returns:
        str: The generated addmodel command
    """
    logger.info("Generating addmodel command")
    command = f"addmodel -datacolumn {add_model_options.datacolumn} -m {add_model_options.mode} "
    command += f"{add_model_options.model_path!s} {add_model_options.ms_path!s}"

    return command


def add_model(
    add_model_options: AddModelOptions, container: Path, remove_datacolumn: bool = False
) -> AddModelOptions:
    """Use the ``addmodel`` program to predict the sky-model visibilities
    from a compatible source list (e.g. ``wsclean -save-source-list``)

    Args:
        add_model_options (AddModelOptions): The set of supported options to be supplied to ``addmodel``
        container (Path): The calibrate container that contains the ``addmodel`` program
        remove_datacolumn (bool, optional): Whether to first remove the ``datacolumn`` specified in ``add_model_options`` before predicting. If False it should be overwritten. Defaults to False.

    Returns:
        AddModelOptions: The options used to run ``addmodel`` (same as input)
    """
    if remove_datacolumn:
        remove_columns_from_ms(
            ms=add_model_options.ms_path, columns_to_remove=add_model_options.datacolumn
        )
    add_model_command = add_model_options_to_command(
        add_model_options=add_model_options
    )

    run_singularity_command(
        image=container,
        command=add_model_command,
        bind_dirs=[add_model_options.ms_path, add_model_options.model_path],
    )

    return add_model_options


def get_parser() -> ArgumentParser:
    parser = ArgumentParser(
        description="Predict a set of model visibilities from a model with addmodel"
    )

    parser.add_argument(
        "calibrate_container", type=Path, help="Path to the container with addmodel"
    )
    parser.add_argument(
        "--remove-datacolumn",
        action="store_true",
        help="Remove the column being predicted into before predicting. See the AddModelOptions.datacolumn option.",
    )

    parser = add_options_to_parser(parser=parser, options_class=AddModelOptions)

    return parser


def cli() -> None:
    parser = get_parser()

    args = parser.parse_args()

    add_model_options = create_options_from_parser(
        parser_namespace=args, options_class=AddModelOptions
    )

    add_model(
        add_model_options=add_model_options,
        container=args.calibrate_container,
        remove_datacolumn=args.remove_datacolumn,
    )


if __name__ == "__main__":
    cli()



"""Testing some wsclean functionality."""

from __future__ import annotations

import os
import shutil
from pathlib import Path
from typing import Any

import pytest

from flint.exceptions import (
    AttemptRerunException,
    CleanDivergenceError,
    NamingException,
)
from flint.imager.wsclean import (
    ImageSet,
    WSCleanOptions,
    WSCleanResult,
    _rename_wsclean_file,
    _rename_wsclean_title,
    _resolve_wsclean_key_value_to_cli_str,
    _wsclean_output_callback,
    combine_image_set_to_cube,
    create_wsclean_cmd,
    create_wsclean_name_argument,
    get_wsclean_output_names,
    get_wsclean_output_source_list_path,
    merge_image_sets,
    rename_wsclean_prefix_in_image_set,
    split_and_get_image_set,
    split_image_set,
)
from flint.logging import logger
from flint.naming import create_imaging_name_prefix
from flint.options import MS
from flint.utils import get_packaged_resource_path


def test_get_wsclean_output_source_list_path():
    """Wsclean can be configured out output a source list of the
    components, their brightness and relative size that were placed
    throughout cleaning. Here we be testing whether we can
    generate the expected name"""

    example = Path("/flint/pirates/SB58992.RACS_1726-73.beam22.ms")
    source_path = Path("/flint/pirates/SB58992.RACS_1726-73.beam22.i-sources.txt")

    test_source_path = get_wsclean_output_source_list_path(name_path=example, pol="i")
    assert source_path == test_source_path

    example = Path("/flint/pirates/SB58992.RACS_1726-73.beam22")
    source_path = Path("/flint/pirates/SB58992.RACS_1726-73.beam22.i-sources.txt")

    test_source_path = get_wsclean_output_source_list_path(name_path=example, pol="i")
    assert source_path == test_source_path

    example = "SB58992.RACS_1726-73.beam22"
    source_path = Path("SB58992.RACS_1726-73.beam22.i-sources.txt")

    test_source_path = get_wsclean_output_source_list_path(name_path=example, pol="i")
    assert source_path == test_source_path

    example = "SB58992.RACS_1726-73.beam22"
    source_path = Path("SB58992.RACS_1726-73.beam22-sources.txt")

    test_source_path = get_wsclean_output_source_list_path(name_path=example, pol=None)
    assert source_path == test_source_path


@pytest.fixture
def ms_example(tmpdir):
    ms_zip = Path(
        get_packaged_resource_path(
            package="flint.data.tests",
            filename="SB39400.RACS_0635-31.beam0.small.ms.zip",
        )
    )
    outpath = Path(tmpdir) / "39400"

    shutil.unpack_archive(ms_zip, outpath)

    ms_path = Path(outpath) / "SB39400.RACS_0635-31.beam0.small.ms"

    return ms_path


@pytest.fixture(scope="session", autouse=True)
def set_env():
    """Set up variables for a specific test"""
    os.environ["LOCALDIR"] = "Pirates/be/here"


def test_rename_wsclean_path_move(tmpdir: Any):
    """Rename the wsclean supplied part of a filename while moving a file"""
    test_path = Path(tmpdir) / "move_file/"
    test_path.mkdir(parents=True, exist_ok=True)

    ex = test_path / Path("SB39400.RACS_0635-31.beam33.poli-MFS-image.fits")
    out_ex = test_path / Path("SB39400.RACS_0635-31.beam33.poli.MFS.image.fits")

    with open(ex, "w") as out_file:
        out_file.write("example")

    assert ex.exists()
    assert not out_ex.exists()
    assert _rename_wsclean_file(input_path=ex, rename_file=True) == out_ex
    assert not ex.exists()
    assert out_ex.exists()


def _write_test_image(items: Any):
    for item in items:
        with Path(item).open("w") as out_file:
            out_file.write(str(item))


def test_rename_wsclean_image_set(tmpdir: Any):
    """Ensure that items described in an image set are able to be properly renamed"""

    test_dir = Path(tmpdir) / "image_setrename"
    test_dir.mkdir(parents=True, exist_ok=True)

    # create some test files and ensure they all exist
    keys: dict[Any, Any] = {}
    prefix = f"{test_dir!s}/SB39400.RACS_0635-31.beam33.i"
    keys["prefix"] = prefix
    for mode in ("image", "residual"):
        items = [
            Path(f"{prefix}-{subband:04d}-{mode}.fits") for subband in range(4)
        ] + [Path(f"{prefix}-MFS-{mode}.fits")]
        _write_test_image(items=items)
        keys[mode] = items
        assert all([Path(f).exists() for f in items])

    # form the image set that will have the wsclean appended properties string renamed
    image_set = ImageSet(**keys)
    assert isinstance(image_set, ImageSet)
    new_image_set = rename_wsclean_prefix_in_image_set(input_image_set=image_set)

    # test to see thhat files exists
    assert new_image_set.prefix == prefix
    assert new_image_set.image is not None
    assert all([file.exists() for file in new_image_set.image])
    assert new_image_set.residual is not None
    assert all([file.exists() for file in new_image_set.residual])

    # and ensure the originals no longer exist
    assert all([not Path(file).exists() for file in keys["image"]])
    assert all([not (file).exists() for file in keys["residual"]])


def test_rename_wsclean_path():
    """Rename the wsclean supplied part of a filename"""

    ex = Path("SB39400.RACS_0635-31.beam33.poli-MFS-image.fits")
    out_ex = Path("SB39400.RACS_0635-31.beam33.poli.MFS.image.fits")
    assert _rename_wsclean_file(input_path=ex) == out_ex

    ex = Path("SB39400.RACS_0635-31.beam33.poli-MFS-image")
    out_ex = Path("SB39400.RACS_0635-31.beam33.poli.MFS.image")
    assert _rename_wsclean_file(input_path=ex) == out_ex

    ex = Path("/a/path/that/is/a/parent/SB39400.RACS_0635-31.beam33.poli-MFS-image")
    out_ex = Path("/a/path/that/is/a/parent/SB39400.RACS_0635-31.beam33.poli.MFS.image")
    assert _rename_wsclean_file(input_path=ex) == out_ex

    # This one makes sure that the second component of ch0234-0345 is not pixked up
    # as the wsclean added subband indicator
    ex = Path("/a/path/that/is/a/parent/SB39400.RACS_0635-31.beam33.poli-MFS-image")
    out_ex = Path("/a/path/that/is/a/parent/SB39400.RACS_0635-31.beam33.poli.MFS.image")
    assert _rename_wsclean_file(input_path=ex) == out_ex

    ex = Path(
        "/a/path/that/is/a/parent/SB57516.RACS_0929-81.beam35.round4.i.ch0287-0288-image.fits"
    )
    out_ex = Path(
        "/a/path/that/is/a/parent/SB57516.RACS_0929-81.beam35.round4.i.ch0287-0288.image.fits"
    )
    assert _rename_wsclean_file(input_path=ex) == out_ex


def test_rename_stokes_v_model():
    """Some model files are not being renamed correctly. Arr"""

    ex = "SB57988.RACS_1415-46.beam34.round4.v-MFS-model.fits"
    out_ex = "SB57988.RACS_1415-46.beam34.round4.v.MFS.model.fits"

    assert _rename_wsclean_title(name_str=ex) == out_ex


def test_regex_rename_wsclean_title():
    """Rename the wsclean supplied using regex"""

    ex = "SB39400.RACS_0635-31.beam33.poli-MFS-image.fits"
    out_ex = "SB39400.RACS_0635-31.beam33.poli.MFS.image.fits"
    assert _rename_wsclean_title(name_str=ex) == out_ex

    ex = "SB39400.RACS_0635-31.beam33.poli-MFS-image"
    out_ex = "SB39400.RACS_0635-31.beam33.poli.MFS.image"
    assert _rename_wsclean_title(name_str=ex) == out_ex

    ex = "SB39400.RACS_0635-31.beam33.poli-MFS-image"
    out_ex = "SB39400.RACS_0635.31.beam33.poli.MFS.image"
    assert not _rename_wsclean_title(name_str=ex) == out_ex

    ex = "SB39400.RACS_0635-31.beam33.poli.MFS.image.fits"
    out_ex = "SB39400.RACS_0635-31.beam33.poli.MFS.image.fits"
    assert _rename_wsclean_title(name_str=ex) == out_ex
    assert _rename_wsclean_title(name_str=ex) is ex

    ex = "SB39400.RACS_0635-31.beam33.poli-i-MFS-image"
    out_ex = "SB39400.RACS_0635-31.beam33.poli.i.MFS.image"
    assert _rename_wsclean_title(name_str=ex) == out_ex

    ex = "SB39400.RACS_0635-31.beam33.ch109-110-i-MFS-image"
    out_ex = "SB39400.RACS_0635-31.beam33.ch109-110.i.MFS.image"
    assert _rename_wsclean_title(name_str=ex) == out_ex

    ex = "SB39400.RACS_0635-31.beam33.i.ch109-110-i-MFS-image"
    out_ex = "SB39400.RACS_0635-31.beam33.i.ch109-110.i.MFS.image"
    assert _rename_wsclean_title(name_str=ex) == out_ex


def test_regex_stokes_wsclean_title():
    """Test whether all stokes values are picked up properly"""

    prefix = "SB39400.RACS_0635-31.beam33.poli."
    end = "-MFS-image.fits"
    transformed = end.replace("-", ".")

    for stokes in ("i", "q", "u", "v", "xx", "xy", "yx", "yy"):
        ex = f"{prefix}-{stokes}{end}"
        out_ex = f"{prefix}.{stokes}{transformed}"
        assert _rename_wsclean_title(name_str=ex) == out_ex

    name = "SB59058.RACS_1626-84.beam34.round4.i.ch287-288-image.fits"
    out_name = "SB59058.RACS_1626-84.beam34.round4.i.ch287-288.image.fits"
    assert _rename_wsclean_title(name_str=name) == out_name


def test_combine_subbands_to_cube(tmpdir):
    """Load in example fits images to combine into a cube"""
    files = [
        get_packaged_resource_path(
            package="flint.data.tests",
            filename=f"SB56659.RACS_0940-04.beam17.round3-000{i}-image.sub.fits",
        )
        for i in range(3)
    ]
    files = [Path(shutil.copy(Path(f), Path(tmpdir))) for f in files]

    assert len(files) == 3
    assert all([f.exists() for f in files])
    file_parent = files[0].parent
    prefix = f"{file_parent}/SB56659.RACS_0940-04.beam17.round3"
    image_set = ImageSet(
        prefix=prefix,
        image=files,
    )

    new_image_set = combine_image_set_to_cube(
        image_set=image_set, remove_original_images=False
    )

    assert new_image_set.prefix == image_set.prefix
    assert len(new_image_set.image) == 1

    with pytest.raises(TypeError):
        _ = combine_image_set_to_cube(image_set=files, remove_original_images=False)  # type: ignore


def test_combine_subbands_to_cube2(tmpdir):
    """Load in example fits images to combine into a cube without deleting original"""
    files = [
        get_packaged_resource_path(
            package="flint.data.tests",
            filename=f"SB56659.RACS_0940-04.beam17.round3-000{i}-image.sub.fits",
        )
        for i in range(3)
    ]
    files = [Path(shutil.copy(Path(f), Path(tmpdir))) for f in files]

    assert len(files) == 3
    assert all([f.exists() for f in files])
    file_parent = files[0].parent
    prefix = f"{file_parent}/SB56659.RACS_0940-04.beam17.round3"
    image_set = ImageSet(
        prefix=prefix,
        image=files,
    )

    new_image_set = combine_image_set_to_cube(
        image_set=image_set, remove_original_images=True
    )
    assert all([not file.exists() for file in files])
    assert new_image_set.prefix == image_set.prefix
    assert len(new_image_set.image) == 1


def test_resolve_key_value_to_cli():
    """The wsclean command generation operates over keys and values, and
    the formatting is partly based on the type a value has. This goes through
    those checks"""
    res = _resolve_wsclean_key_value_to_cli_str("size", 1024)
    assert res.cmd == "-size 1024 1024"
    assert res.bindpath is None
    assert res.unknown is None

    res = _resolve_wsclean_key_value_to_cli_str("no_update_model_required", True)
    assert res.cmd == "-no-update-model-required"
    assert res.bindpath is None
    assert res.unknown is None

    res = _resolve_wsclean_key_value_to_cli_str("no_update_model_required", False)
    assert res.cmd is None
    assert res.bindpath is None
    assert res.unknown is None

    res = _resolve_wsclean_key_value_to_cli_str("temp_dir", Path("jack/sparrow"))
    assert res.cmd == "-temp-dir jack/sparrow"
    assert res.bindpath == Path("jack/sparrow")
    assert res.unknown is None

    unknown = WSCleanOptions
    res = _resolve_wsclean_key_value_to_cli_str("temp_dir", unknown)
    assert res.cmd is None
    assert res.bindpath is None
    assert res.unknown == ("temp_dir", unknown)

    ignore = WSCleanOptions
    res = _resolve_wsclean_key_value_to_cli_str("flint_this_should_be_ignored", ignore)
    assert res.cmd is None
    assert res.bindpath is None
    assert res.unknown is None
    assert res.ignore


def test_create_wsclean_name(ms_example):
    """Test the creation of a wsclean name argument"""
    name = create_imaging_name_prefix(ms_path=ms_example)
    assert name == "SB39400.RACS_0635-31.beam0.small"

    for pol in ("i", "I"):
        name = create_imaging_name_prefix(ms_path=ms_example, pol=pol)
        assert name == "SB39400.RACS_0635-31.beam0.small.i"


def test_create_wsclean_name_argument(ms_example):
    """Ensure that the generated name argument behaves as expected"""

    ms = MS.cast(ms=Path(ms_example))
    wsclean_options = WSCleanOptions()
    name_argument_path = create_wsclean_name_argument(
        wsclean_options=wsclean_options, ms=ms
    )

    parent = str(Path(ms_example).parent)
    assert isinstance(name_argument_path, Path)
    assert f"{parent}/SB39400.RACS_0635-31.beam0.small.i" == str(name_argument_path)

    wsclean_options_2 = WSCleanOptions(temp_dir="/jack/sparrow")
    name_argument_path = create_wsclean_name_argument(
        wsclean_options=wsclean_options_2, ms=ms
    )

    assert "/jack/sparrow/SB39400.RACS_0635-31.beam0.small.i" == str(name_argument_path)


def test_create_wsclean_command(ms_example):
    """Test whether WSCleanOptions can be correctly cast to a command string"""
    wsclean_options = WSCleanOptions()

    command = create_wsclean_cmd(
        ms=MS.cast(ms_example), wsclean_options=wsclean_options
    )
    assert isinstance(command, WSCleanResult)


def test_create_wsclean_command_with_environment(ms_example):
    """Test whether WSCleanOptions can be correctly cast to a command string"""
    wsclean_options = WSCleanOptions(temp_dir="$LOCALDIR")

    command = create_wsclean_cmd(
        ms=MS.cast(ms_example), wsclean_options=wsclean_options
    )
    assert isinstance(command, WSCleanResult)
    assert "Pirates/be/here" in command.cmd
    assert command.cmd.startswith("wsclean ")


def test_wsclean_divergence():
    """Make sure the wsclean call back function picks up divergence and raises appropriate errors"""
    good = (
        "Iteration 59228, scale 0 px : -862.94 µJy at 3729,3746",
        "Opening reordered part 0 spw 0 for /scratch3/gal16b/flint_peel/40470/SB40470.RACS_1237+00.beam4.round1.ms",
        "Opening reordered part 0 spw 0 for /scratch3/gal16b/flint_peel/40470/SB40470.RACS_1237+00.beam4.round1.ms",
        "Although KJy there is no iterate ion, not the lack of a capital-I and the space, clever pirate",
    )
    for g in good:
        _wsclean_output_callback(line=g)

    bad = "Iteration 59228, scale 0 px : -862.94 KJy at 3729,3746"
    with pytest.raises(CleanDivergenceError):
        _wsclean_output_callback(line=bad)

    with pytest.raises(AssertionError):
        _wsclean_output_callback(line=tuple("A tuple of text".split()))


def test_attemptrerun_wsclean_output_callback():
    """Some known lines output by wsclean can be caused by some transient
    type of error. In such a situation AttemptRerunException should
    be raised."""

    good = (
        "Iteration 59228, scale 0 px : -862.94 µJy at 3729,3746",
        "Opening reordered part 0 spw 0 for /scratch3/gal16b/flint_peel/40470/SB40470.RACS_1237+00.beam4.round1.ms",
        "Opening reordered part 0 spw 0 for /scratch3/gal16b/flint_peel/40470/SB40470.RACS_1237+00.beam4.round1.ms",
        "Although Input/output is here, it is not next to error",
        "Similar with temporary data file error opening error",
    )
    for g in good:
        _wsclean_output_callback(line=g)

    bad = (
        "Input/output error",
        "But why is the rum gone... Input/output error",
        "Input/output error should cause a remake of Pirates of the Caribbean",
    )
    for b in bad:
        with pytest.raises(AttemptRerunException):
            _wsclean_output_callback(line=b)


def test_wsclean_output_named_raises():
    with pytest.raises(FileExistsError):
        _ = get_wsclean_output_names(
            prefix="JackSparrow", subbands=4, verify_exists=True
        )


def test_wsclean_output_named_check_when_adding():
    image_set = get_wsclean_output_names(
        prefix="JackSparrow",
        subbands=4,
        verify_exists=True,
        check_exists_when_adding=True,
    )

    assert isinstance(image_set, ImageSet)
    assert len(image_set.image) == 0


def test_wsclean_output_named():
    image_set = get_wsclean_output_names(prefix="JackSparrow", subbands=4)

    assert isinstance(image_set, ImageSet)
    assert image_set.prefix == "JackSparrow"

    assert image_set.image is not None
    assert len(image_set.image) == 5
    assert isinstance(image_set.image[0], Path)

    assert image_set.dirty is not None
    assert len(image_set.dirty) == 5
    assert isinstance(image_set.dirty[0], Path)

    assert image_set.model is not None
    assert len(image_set.model) == 5
    assert isinstance(image_set.model[0], Path)

    assert image_set.residual is not None
    assert len(image_set.residual) == 5
    assert isinstance(image_set.residual[0], Path)

    assert image_set.psf is not None
    assert len(image_set.psf) == 5
    assert isinstance(image_set.psf[0], Path)


def test_wsclean_output_named_wpols():
    image_set = get_wsclean_output_names(
        prefix="JackSparrow", subbands=4, pols=("I", "Q")
    )

    assert isinstance(image_set, ImageSet)
    assert image_set.prefix == "JackSparrow"

    expected = 10
    assert image_set.image is not None
    assert len(image_set.image) == expected
    assert isinstance(image_set.image[0], Path)

    assert image_set.dirty is not None
    assert len(image_set.dirty) == expected
    assert isinstance(image_set.dirty[0], Path)

    assert image_set.model is not None
    assert len(image_set.model) == expected
    assert isinstance(image_set.model[0], Path)

    assert image_set.residual is not None
    assert len(image_set.residual) == expected
    assert isinstance(image_set.residual[0], Path)

    assert image_set.psf is not None
    assert len(image_set.psf) == 5  # PSF is the same across all pols
    assert isinstance(image_set.psf[0], Path)

    assert image_set.image[0] == Path("JackSparrow-0000-I-image.fits")
    assert image_set.image[4] == Path("JackSparrow-MFS-I-image.fits")
    assert image_set.image[5] == Path("JackSparrow-0000-Q-image.fits")
    assert image_set.image[9] == Path("JackSparrow-MFS-Q-image.fits")


def test_wsclean_output_named_nomfs():
    image_set = get_wsclean_output_names(
        prefix="JackSparrow", subbands=4, include_mfs=False
    )

    assert isinstance(image_set, ImageSet)
    assert image_set.prefix == "JackSparrow"

    assert image_set.image is not None
    assert len(image_set.image) == 4
    assert isinstance(image_set.image[0], Path)

    assert image_set.dirty is not None
    assert len(image_set.dirty) == 4
    assert isinstance(image_set.dirty[0], Path)

    assert image_set.model is not None
    assert len(image_set.model) == 4
    assert isinstance(image_set.model[0], Path)

    assert image_set.residual is not None
    assert len(image_set.residual) == 4
    assert isinstance(image_set.residual[0], Path)

    assert image_set.psf is not None
    assert len(image_set.psf) == 4
    assert isinstance(image_set.psf[0], Path)


def test_wsclean_names_no_subbands():
    """The spectral line modes image per channel, so there is therefore no subband type
    in the wsclean named output"""
    image_set = get_wsclean_output_names(
        prefix="JackSparrow", subbands=1, include_mfs=False
    )

    assert isinstance(image_set, ImageSet)
    assert image_set.prefix == "JackSparrow"

    assert image_set.image
    assert len(image_set.image) == 1
    assert image_set.image[0] == Path("JackSparrow-image.fits")

    assert image_set.psf
    assert len(image_set.psf) == 1
    assert image_set.psf[0] == Path("JackSparrow-psf.fits")


def test_merge_image_sets():
    """Test merging image sets"""
    image_set = get_wsclean_output_names(
        prefix="JackSparrow", subbands=4, include_mfs=False
    )

    image_set2 = get_wsclean_output_names(
        prefix="JackSparrow", subbands=4, include_mfs=False
    )

    merged = merge_image_sets(image_sets=[image_set, image_set2])

    assert isinstance(merged, ImageSet)
    assert merged.prefix == "JackSparrow"

    assert merged.image is not None
    assert len(merged.image) == 8
    assert isinstance(merged.image[0], Path)

    assert merged.dirty is not None
    assert len(merged.dirty) == 8
    assert isinstance(merged.dirty[0], Path)

    assert merged.model is not None
    assert len(merged.model) == 8
    assert isinstance(merged.model[0], Path)

    assert merged.residual is not None
    assert len(merged.residual) == 8
    assert isinstance(merged.residual[0], Path)

    assert merged.psf is not None
    assert len(merged.psf) == 8
    assert isinstance(merged.psf[0], Path)


def test_split_image_set():
    ms = Path("SB1234.FieldNme.beam00.round4.ms")
    i_prefix = create_imaging_name_prefix(ms, pol="i")
    q_prefix = create_imaging_name_prefix(ms, pol="q")
    u_prefix = create_imaging_name_prefix(ms, pol="u")
    v_prefix = create_imaging_name_prefix(ms, pol="v")

    i_image_set = get_wsclean_output_names(
        prefix=i_prefix,
        subbands=4,
        include_mfs=False,
        pols="I",
    )
    q_image_set = get_wsclean_output_names(
        prefix=q_prefix,
        subbands=4,
        include_mfs=False,
        pols="Q",
    )
    u_image_set = get_wsclean_output_names(
        prefix=u_prefix,
        subbands=4,
        include_mfs=False,
        pols="U",
    )
    v_image_set = get_wsclean_output_names(
        prefix=v_prefix,
        subbands=4,
        include_mfs=False,
        pols="V",
    )

    image_set = ImageSet(
        prefix=i_prefix,
        image=i_image_set.image
        + q_image_set.image
        + u_image_set.image
        + v_image_set.image,
        dirty=i_image_set.dirty
        + q_image_set.dirty
        + u_image_set.dirty
        + v_image_set.dirty,
        model=i_image_set.model
        + q_image_set.model
        + u_image_set.model
        + v_image_set.model,
        residual=i_image_set.residual
        + q_image_set.residual
        + u_image_set.residual
        + v_image_set.residual,
        psf=i_image_set.psf + q_image_set.psf + u_image_set.psf + v_image_set.psf,
    )

    split_dict = split_image_set(image_set=image_set, by="pol", mode="image")

    logger.info(f"Split dict: {split_dict}")

    assert isinstance(split_dict, dict)
    for s, (key, val) in zip(
        ("i", "q", "u", "v"),
        split_dict.items(),
    ):
        assert isinstance(val, list)
        assert len(val) == 4
        assert isinstance(val[0], Path)
        assert key == s

    with pytest.raises(NamingException):
        split_image_set(image_set=image_set, by="pol", mode="plunder")


def test_split_and_get_image_set():
    ms = Path("SB1234.FieldNme.beam00.round4.ms")
    i_prefix = create_imaging_name_prefix(ms, pol="i")
    q_prefix = create_imaging_name_prefix(ms, pol="q")
    u_prefix = create_imaging_name_prefix(ms, pol="u")
    v_prefix = create_imaging_name_prefix(ms, pol="v")

    i_image_set = get_wsclean_output_names(
        prefix=i_prefix,
        subbands=4,
        include_mfs=False,
        pols="I",
    )
    q_image_set = get_wsclean_output_names(
        prefix=q_prefix,
        subbands=4,
        include_mfs=False,
        pols="Q",
    )
    u_image_set = get_wsclean_output_names(
        prefix=u_prefix,
        subbands=4,
        include_mfs=False,
        pols="U",
    )
    v_image_set = get_wsclean_output_names(
        prefix=v_prefix,
        subbands=4,
        include_mfs=False,
        pols="V",
    )

    image_set = ImageSet(
        prefix=i_prefix,
        image=i_image_set.image
        + q_image_set.image
        + u_image_set.image
        + v_image_set.image,
        dirty=i_image_set.dirty
        + q_image_set.dirty
        + u_image_set.dirty
        + v_image_set.dirty,
        model=i_image_set.model
        + q_image_set.model
        + u_image_set.model
        + v_image_set.model,
        residual=i_image_set.residual
        + q_image_set.residual
        + u_image_set.residual
        + v_image_set.residual,
        psf=i_image_set.psf + q_image_set.psf + u_image_set.psf + v_image_set.psf,
    )

    split_list = split_and_get_image_set(
        image_set=image_set, get="i", by="pol", mode="image"
    )

    assert isinstance(split_list, list)
    assert len(split_list) == 4
    assert isinstance(split_list[0], Path)
    assert split_list[0].name.endswith("I-image.fits")



"""Small tests for items related to measurement sets
and the MS class
"""

from __future__ import annotations

import shutil
from pathlib import Path

import numpy as np
import pytest
from casacore.tables import table
from pydantic import ValidationError

from flint.calibrate.aocalibrate import ApplySolutions
from flint.ms import (
    MS,
    check_column_in_ms,
    consistent_channelwise_frequencies,
    copy_and_preprocess_casda_askap_ms,
    find_mss,
    get_phase_dir_from_ms,
    remove_columns_from_ms,
    rename_ms_and_columns_for_selfcal,
    subtract_model_from_data_column,
)
from flint.utils import get_packaged_resource_path


def test_consistent_channelwise_frequencies():
    """Some steps the channels in a set of MSs all need
    to be the same, in the same relative order"""

    # Make up some floating point numbners
    freqs = [np.arange(100) * np.pi * 5 for _ in list(range(46))]
    result = consistent_channelwise_frequencies(freqs=freqs)
    assert np.all(result)

    result = consistent_channelwise_frequencies(freqs=np.array(freqs))
    assert np.all(result)

    freqs[10][4] = 4444444.0
    result = consistent_channelwise_frequencies(freqs=np.array(freqs))
    assert not result[10]
    assert np.all(result[11:])
    assert np.all(result[:10])


def test_find_mss(tmpdir):
    """Make sure that the glob finding of the MSs can actually find
    the MSs. The expected count check is also assessed"""
    tmpdir = Path(tmpdir)
    for name in range(45):
        new_ms = tmpdir / f"SB1234.Pirate_1234+456.beam{name}.ms"
        new_ms.mkdir()

        new_folder = tmpdir / f"not_and_ms_{name}.folder"
        new_folder.mkdir()

    res = find_mss(mss_parent_path=tmpdir, expected_ms_count=45)
    assert len(res) == 45

    with pytest.raises(AssertionError):
        _ = find_mss(mss_parent_path=tmpdir, expected_ms_count=49005)


def test_find_mss_withdatacolumn(tmpdir):
    """Same as above but with setting a data column"""
    tmpdir = Path(tmpdir) / "Another_Pirate"
    for name in range(45):
        new_ms: Path = tmpdir / f"SB1234.Pirate_1234+456.beam{name}.ms"
        new_ms.mkdir(parents=True, exist_ok=True)

        new_folder = tmpdir / f"not_and_ms_{name}.folder"
        new_folder.mkdir()

    res = find_mss(
        mss_parent_path=tmpdir, expected_ms_count=45, data_column="BlackBeard"
    )
    assert len(res) == 45
    assert all([ms.column == "BlackBeard" for ms in res])

    with pytest.raises(AssertionError):
        _ = find_mss(mss_parent_path=tmpdir, expected_ms_count=49005)


@pytest.fixture
def casda_example(tmpdir):
    ms_zip = Path(
        get_packaged_resource_path(
            package="flint.data.tests",
            filename="scienceData.EMU_0529-60.SB50538.EMU_0529-60.beam08_averaged_cal.leakage.ms.zip",
        )
    )
    outpath = Path(tmpdir) / "extract"

    shutil.unpack_archive(ms_zip, outpath)

    ms_path = (
        Path(outpath)
        / "scienceData.EMU_0529-60.SB50538.EMU_0529-60.beam08_averaged_cal.leakage.ms"
    )

    return ms_path


def _test_the_data(ms):
    """Some very simple tests for the rotation. The expected numbers come from manually
    stabbing the MSs"""
    from casacore.tables import table

    with table(str(ms), ack=False) as tab:
        data = tab.getcol("DATA")
        inst_data = tab.getcol("INSTRUMENT_DATA")
        colnames = tab.colnames()

    assert all([col in colnames for col in ("DATA", "INSTRUMENT_DATA")])

    expected_inst_data = np.array(
        [5.131794 - 23.130766j, 45.26275 - 45.140232j, 0.80312335 + 0.41873842j],
        dtype=np.complex64,
    )
    assert np.allclose(inst_data[:, 10, 0], expected_inst_data)

    expected_data = np.array(
        [-12.364758 - 59.172283j, -10.334289 - 97.017624j, 1.022179 - 0.18529199j],
        dtype=np.complex64,
    )
    assert np.allclose(data[:, 10, 0], expected_data)


def test_copy_preprocess_ms(casda_example, tmpdir):
    """Run the copying and preprocessing for the casda askap. This is not testing the actual contents or the
    output visibility file yet. Just sanity around the process."""

    output_path = Path(tmpdir) / "casda_ms"

    new_ms = copy_and_preprocess_casda_askap_ms(
        casda_ms=Path(casda_example), output_directory=output_path
    )
    _test_the_data(ms=new_ms.path)

    # When file format not recognised
    with pytest.raises(ValueError):
        copy_and_preprocess_casda_askap_ms(
            casda_ms=Path(casda_example) / "Thisdoesnotexist",
            output_directory=output_path,
        )

    # When input directory does not exist
    with pytest.raises(AssertionError):
        copy_and_preprocess_casda_askap_ms(
            casda_ms=Path(
                "thisdoesnotexist/scienceData.EMU_0529-60.SB50538.EMU_0529-60.beam08_averaged_cal.leakage.ms"
            ),
            output_directory=output_path / "New",
        )


@pytest.fixture
def ms_example(tmpdir):
    ms_zip = Path(
        get_packaged_resource_path(
            package="flint.data.tests",
            filename="SB39400.RACS_0635-31.beam0.small.ms.zip",
        )
    )
    outpath = Path(tmpdir) / "39400"

    shutil.unpack_archive(ms_zip, outpath)

    ms_path = Path(outpath) / "SB39400.RACS_0635-31.beam0.small.ms"

    return ms_path


def test_check_column_in_ms(ms_example):
    """See whether columns are present in the MS, and whether the order of checking is correct"""
    ms = MS(path=Path(ms_example), column="DATA")

    assert check_column_in_ms(ms=ms)
    assert not check_column_in_ms(ms=ms, column="NoExists")

    with pytest.raises(ValueError):
        check_column_in_ms(ms=ms.with_options(column=None))


def _get_columns(ms_path):
    with table(str(ms_path), readonly=True, ack=False) as tab:
        return tab.colnames()


def test_rename_ms_and_columns_for_selfcal_correct2data(ms_example, tmpdir):
    """Sanity around renaming a MS and handling the columns that should be renamed"""
    ms = MS.cast(Path(ms_example))
    with table(str(ms.path), readonly=False, ack=False) as tab:
        tab.renamecol("DATA", "CORRECTED_DATA")

    colnames = _get_columns(ms_path=ms.path)

    assert "DATA" not in colnames
    assert "CORRECTED_DATA" in colnames

    target = Path(tmpdir) / "target_directory"
    target.mkdir(parents=True)
    target = target / ms.path.name

    new_ms = rename_ms_and_columns_for_selfcal(ms=ms, target=target)
    new_colnames = _get_columns(ms_path=new_ms.path)

    assert new_ms.path == target
    assert "DATA" in new_colnames
    assert "CORRECTED_DATA" not in new_colnames


def test_rename_ms_and_columns_for_selfcal(ms_example, tmpdir):
    """Sanity around renaming a MS and handling the columns that should be renamed"""
    ms = MS.cast(Path(ms_example))
    colnames = _get_columns(ms_path=ms.path)

    assert "DATA" in colnames

    target = Path(tmpdir) / "target_directory"
    target.mkdir(parents=True)
    target = target / ms.path.name

    new_ms = rename_ms_and_columns_for_selfcal(ms=ms, target=target)
    new_colnames = _get_columns(ms_path=new_ms.path)

    assert new_ms.path == target
    assert "DATA" in new_colnames


def test_phase_dir(ms_example):
    pos = get_phase_dir_from_ms(ms=ms_example)

    assert np.isclose(pos.ra.deg, 98.211959)
    assert np.isclose(pos.dec.deg, -30.86099889)


def test_ms_self_attribute():
    ex = Path("example/jack_sparrow.ms")
    ms = MS(path=ex)

    assert ms.ms.path == ex


def test_ms_from_options():
    path = Path("example.ms")
    solutions = ApplySolutions(
        cmd="none", solution_path=Path("example_sols.bin"), ms=MS(path=path)
    )

    example_ms = MS.cast(solutions)
    ms = MS(path=path)

    assert isinstance(example_ms, MS)
    assert example_ms == ms


def test_raise_error_ms_from_options():
    """Ensure validation error is raised when not casting correct
    ms path type"""
    # TODO: remove this test as pydantic prevents this from happening
    path = Path("example.ms")
    with pytest.raises(ValidationError):
        _ = ApplySolutions(cmd="none", solution_path=Path("example_sols.bin"), ms=path)

    # with pytest.raises((MSError, ValidationError)):
    #     MS.cast(solutions)


@pytest.fixture
def ms_remove_example(tmpdir):
    """Create a copy of the MS that will be modified."""
    ms_zip = Path(
        get_packaged_resource_path(
            package="flint.data.tests",
            filename="SB39400.RACS_0635-31.beam0.small.ms.zip",
        )
    )
    outpath = Path(tmpdir) / "Removecols_39400"

    shutil.unpack_archive(ms_zip, outpath)

    ms_path = Path(outpath) / "SB39400.RACS_0635-31.beam0.small.ms"

    return ms_path


def _get_column_names(ms_path):
    with table(str(ms_path)) as tab:
        column_names = tab.colnames()

    return column_names


def test_remove_columns_from_ms(ms_remove_example):
    """Load an example MS to remove columns from"""
    original_columns = _get_column_names(ms_path=ms_remove_example)

    removed_columns = remove_columns_from_ms(
        ms=ms_remove_example, columns_to_remove="DATA"
    )

    updated_columns = _get_column_names(ms_path=ms_remove_example)
    diff = set(original_columns) - set(updated_columns)
    assert len(diff) == 1
    assert next(iter(diff)) == "DATA"
    assert removed_columns[0] == "DATA"
    assert len(removed_columns) == 1

    removed_columns = remove_columns_from_ms(
        ms=ms_remove_example, columns_to_remove="DATA"
    )
    assert len(removed_columns) == 0


@pytest.fixture
def casda_taql_example(tmpdir):
    ms_zip = Path(
        get_packaged_resource_path(
            package="flint.data.tests",
            filename="scienceData.EMU_0529-60.SB50538.EMU_0529-60.beam08_averaged_cal.leakage.ms.zip",
        )
    )
    outpath = Path(tmpdir) / "taqlsubtract"

    shutil.unpack_archive(ms_zip, outpath)

    ms_path = (
        Path(outpath)
        / "scienceData.EMU_0529-60.SB50538.EMU_0529-60.beam08_averaged_cal.leakage.ms"
    )

    return ms_path


def test_subtract_model_from_data_column(casda_taql_example):
    """Ensure we can subtact the model from the data via taql"""
    ms = Path(casda_taql_example)
    assert ms.exists()
    ms = MS(path=ms)

    from casacore.tables import makearrcoldesc, maketabdesc

    with table(str(ms.path), readonly=False) as tab:
        data = tab.getcol("DATA")
        ones = np.ones_like(data, dtype=data.dtype)

        tab.putcol(columnname="DATA", value=ones)

        if "MODEL_DATA" not in tab.colnames():
            coldesc = tab.getdminfo("DATA")
            coldesc["NAME"] = "MODEL_DATA"
            tab.addcols(
                maketabdesc(makearrcoldesc("MODEL_DATA", 0.0 + 0j, ndim=2)), coldesc
            )
            tab.flush()
        tab.putcol(columnname="MODEL_DATA", value=ones)
        tab.flush()

    ms = subtract_model_from_data_column(
        ms=ms, model_column="MODEL_DATA", data_column="DATA"
    )
    with table(str(ms.path)) as tab:
        data = tab.getcol("DATA")
        assert np.all(data == 0 + 0j)


def test_subtract_model_from_data_column_ms_column(tmpdir):
    """Ensure we can subtact the model from the data via taql"""
    ms_zip = Path(
        get_packaged_resource_path(
            package="flint.data.tests",
            filename="scienceData.EMU_0529-60.SB50538.EMU_0529-60.beam08_averaged_cal.leakage.ms.zip",
        )
    )
    outpath = Path(tmpdir) / "taqlsubtract2"

    shutil.unpack_archive(ms_zip, outpath)

    ms_path = (
        Path(outpath)
        / "scienceData.EMU_0529-60.SB50538.EMU_0529-60.beam08_averaged_cal.leakage.ms"
    )

    ms = Path(ms_path)
    assert ms.exists()
    ms = MS(path=ms, column="DATA")

    from casacore.tables import makearrcoldesc, maketabdesc

    with table(str(ms.path), readonly=False) as tab:
        data = tab.getcol("DATA")
        ones = np.ones_like(data, dtype=data.dtype)

        tab.putcol(columnname="DATA", value=ones)

        if "MODEL_DATA" not in tab.colnames():
            coldesc = tab.getdminfo("DATA")
            coldesc["NAME"] = "MODEL_DATA"
            tab.addcols(
                maketabdesc(makearrcoldesc("MODEL_DATA", 0.0 + 0j, ndim=2)), coldesc
            )
            tab.flush()
        tab.putcol(columnname="MODEL_DATA", value=ones)
        tab.flush()

    ms = subtract_model_from_data_column(ms=ms, model_column="MODEL_DATA")
    with table(str(ms.path)) as tab:
        data = tab.getcol("DATA")
        assert np.all(data == 0 + 0j)


def test_subtract_model_from_data_column_ms_column_new_column(tmpdir):
    """Ensure we can subtact the model from the data via taql. This test will
    add a new column via taql and will ensure the result is as expected.

    >>> NEW_COLUMN=DATA-MODEL_DATA
    """
    ms_zip = Path(
        get_packaged_resource_path(
            package="flint.data.tests",
            filename="scienceData.EMU_0529-60.SB50538.EMU_0529-60.beam08_averaged_cal.leakage.ms.zip",
        )
    )
    outpath = Path(tmpdir) / "taqlsubtract3"

    shutil.unpack_archive(ms_zip, outpath)

    ms_path = (
        Path(outpath)
        / "scienceData.EMU_0529-60.SB50538.EMU_0529-60.beam08_averaged_cal.leakage.ms"
    )

    ms = Path(ms_path)
    assert ms.exists()
    ms = MS(path=ms, column="DATA")

    from casacore.tables import makearrcoldesc, maketabdesc

    with table(str(ms.path), readonly=False) as tab:
        data = tab.getcol("DATA")
        ones = np.ones_like(data, dtype=data.dtype)

        tab.putcol(columnname="DATA", value=ones)

        if "MODEL_DATA" not in tab.colnames():
            coldesc = tab.getdminfo("DATA")
            coldesc["NAME"] = "MODEL_DATA"
            tab.addcols(
                maketabdesc(makearrcoldesc("MODEL_DATA", 0.0 + 0j, ndim=2)), coldesc
            )
            tab.flush()
        tab.putcol(columnname="MODEL_DATA", value=ones)
        tab.flush()

        colnames = tab.colnames()

    assert "NEW_COLUMN" not in colnames, "Column already exists"

    ms = subtract_model_from_data_column(
        ms=ms, model_column="MODEL_DATA", data_column="DATA", output_column="NEW_COLUMN"
    )
    with table(str(ms.path)) as tab:
        data = tab.getcol("NEW_COLUMN")
        assert np.all(data == 0 + 0j)

        data = tab.getcol("DATA")
        assert np.all(data == ones)



"""Utility functions to make image based masks from images, with the initial
thought being towards FITS images.
"""

from __future__ import annotations

from argparse import ArgumentParser
from pathlib import Path
from typing import Collection, NamedTuple

import astropy.units as u
import numpy as np
from astropy.io import fits
from astropy.wcs import WCS
from radio_beam import Beam
from reproject import reproject_interp
from scipy.ndimage import (
    binary_dilation as scipy_binary_dilation,  # Rename to distinguish from skimage
)
from scipy.ndimage import (
    binary_erosion as scipy_binary_erosion,  # Rename to distinguish from skimage
)
from scipy.ndimage import label, minimum_filter
from scipy.signal import fftconvolve

from flint.logging import logger
from flint.naming import FITSMaskNames, create_fits_mask_names
from flint.options import BaseOptions, add_options_to_parser, create_options_from_parser
from flint.utils import get_pixels_per_beam

# TODO: Need to remove a fair amount of old approaches, and deprecate some of the toy functions


class MaskingOptions(BaseOptions):
    """Contains options for the creation of clean masks from some subject
    image. Clipping levels specified are in units of RMS (or sigma). They
    are NOT in absolute units.
    """

    base_snr_clip: float = 4
    """A base clipping level to be used should other options not be activated"""
    flood_fill: bool = False
    """Whether to attempt to flood fill when constructing a mask. This should be `True` for ``grow_low_snr_islands`` and ``suppress_artefacts`` to have an effect. """
    flood_fill_positive_seed_clip: float = 4.5
    """The clipping level to seed islands that will be grown to lower signal metric"""
    flood_fill_positive_flood_clip: float = 1.5
    """Clipping level used to grow seeded islands down to"""
    flood_fill_use_mbc: bool = False
    """If True, the clipping levels are used as the `increase_factor` when using a minimum absolute clip"""
    flood_fill_use_mbc_box_size: int = 75
    """The size of the mbc box size should mbc be used"""
    flood_fill_use_mbc_adaptive_step_factor: float = 2.0
    """Stepping size used to increase box by should adaptive detect poor boxcar statistics"""
    flood_fill_use_mbc_adaptive_skew_delta: float = 0.2
    """A box is consider too small for a pixel if the fractional proportion of positive pixels is larger than the deviation away of (0.5 + frac). This threshold is therefore 0 to 0.5"""
    flood_fill_use_mbc_adaptive_max_depth: int | None = None
    """Determines the number of adaptive boxcar scales to use when constructing seed mask. If None no adaptive boxcar sizes"""
    grow_low_snr_island: bool = False
    """Whether to attempt to grow a mask to capture islands of low SNR (e.g. diffuse emission)"""
    grow_low_snr_island_clip: float = 1.75
    """The minimum significance levels of pixels to be to seed low SNR islands for consideration"""
    grow_low_snr_island_size: int = 768
    """The number of pixels an island has to be for it to be accepted"""
    beam_shape_erode: bool = False
    """Erode the mask using the shape of the restoring beam"""
    beam_shape_erode_minimum_response: float = 0.6
    """The minimum response of the beam that is used to form the erode structure shape"""


def consider_beam_mask_round(
    current_round: int,
    mask_rounds: str | Collection[int] | int,
    allow_beam_masks: bool = True,
) -> bool:
    """Evaluate whether a self-calibration round should have a beam clean mask
    constructed. Rules are:

    - if `mask_rounds` is a string and is "all", all rounds will have a beam mask
    - if 'mask_rounds' is a single integer, so long as `current_round` is larger it will have a beam mask
    - if `mask_rounds` is iterable and contains `current_round` it will have a beam mask
    - if `allow_beam_masks` is False a False is returned. Otherwise options above are considered.

    Args:
        current_round (int): The current self-calibration round that is being performed
        mask_rounds (Union[str, Collection[int], int]): The rules to consider whether a beam mask is needed
        allow_beam_masks (bool, optional): A global allow / deny. This should be `True` for other rules to be considered. Defaults to True.

    Returns:
        bool: Whether per beam mask should be performed
    """
    logger.info(f"Considering {current_round=} {mask_rounds=} {allow_beam_masks=}")

    # The return below was getting silly to meantally parse
    if not allow_beam_masks:
        return False

    return mask_rounds is not None and (
        (isinstance(mask_rounds, str) and mask_rounds.lower() == "all")
        or (isinstance(mask_rounds, int) and current_round >= mask_rounds)
        or ((isinstance(mask_rounds, (list, tuple))) and current_round in mask_rounds)
    )  # type: ignore


def create_beam_mask_kernel(
    fits_header: fits.Header, kernel_size=100, minimum_response: float = 0.6
) -> np.ndarray:
    """Make a mask using the shape of a beam in a FITS Header object. The
    beam properties in the ehader are used to generate the two-dimensional
    Gaussian main lobe, from which a cut is made based on the minimum
    power.

    Args:
        fits_header (fits.Header): The FITS header to create beam from
        kernel_size (int, optional): Size of the output kernel in pixels. Will be a square. Defaults to 100.
        minimum_response (float, optional): Minimum response of the beam shape for the mask to be constructed from. Defaults to 0.6.

    Raises:
        KeyError: Raised if CDELT1 and CDELT2 missing

    Returns:
        np.ndarray: Boolean mask of the kernel shape
    """
    assert 0.0 < minimum_response < 1.0, (
        f"{minimum_response=}, should be between 0 to 1 (exclusive)"
    )

    POSITION_KEYS = ("CDELT1", "CDELT2")
    if not all([key in fits_header for key in POSITION_KEYS]):
        raise KeyError(f"{POSITION_KEYS=}  all need to be present")

    beam = Beam.from_fits_header(fits_header)
    assert isinstance(beam, Beam)

    cdelt1, cdelt2 = np.abs(fits_header["CDELT1"]), np.abs(fits_header["CDELT2"])  # type: ignore
    assert np.isclose(cdelt1, cdelt2), (
        f"Pixel scales {cdelt1=} {cdelt2=}, but must be equal"
    )

    k = beam.as_kernel(
        pixscale=cdelt1 * u.Unit("deg"), x_size=kernel_size, y_size=kernel_size
    )

    return k.array > (np.max(k.array) * minimum_response)


def beam_shape_erode(
    mask: np.ndarray, fits_header: fits.Header, minimum_response: float = 0.6
) -> np.ndarray:
    """Construct a kernel representing the shape of the restoring beam at
    a particular level, and use it as the basis of a binary erosion of the
    input mask.

    The ``fits_header`` is used to construct the beam shape that matches the
    same pixel size

    Args:
        mask (np.ndarray): The current mask that will be eroded based on the beam shape
        fits_header (fits.Header): The fits header of the mask used to generate the beam kernel shape
        minimum_response (float, optional): The minimum response of the main restoring beam to craft the shape from. Defaults to 0.6.

    Returns:
        np.ndarray: The eroded beam shape
    """

    if not all([key in fits_header for key in ["BMAJ", "BMIN", "BPA"]]):
        logger.warning(
            "Beam parameters missing. Not performing the beam shape erosion. "
        )
        return mask

    logger.info(f"Eroding the mask using the beam shape with {minimum_response=}")
    beam_mask_kernel = create_beam_mask_kernel(
        fits_header=fits_header, minimum_response=minimum_response
    )

    # This handles any unsqueezed dimensions
    beam_mask_kernel = beam_mask_kernel.reshape(
        mask.shape[:-2] + beam_mask_kernel.shape
    )

    erode_mask = scipy_binary_erosion(
        input=mask, iterations=1, structure=beam_mask_kernel
    )

    return erode_mask.astype(mask.dtype)


def extract_beam_mask_from_mosaic(
    fits_beam_image_path: Path, fits_mosaic_mask_names: FITSMaskNames
) -> FITSMaskNames:
    """Extract a region based on an existing FITS image from a masked FITS
    image. Here a masked FITS image is intended to be one created by
    ``create_snr_mask_from_fits``.

    A simple cutout (e.g. ``CutOut2D``) might not work as intended, as
    when creating the field image (the intended use case here) might be
    regridded onto a unique pixel grid that does not correspond to one
    that would be constructed by wsclean.

    Args:
        fits_beam_image_path (Path): The template image with a valid WCS. This region is used to extract the masked region
        fits_mosaic_mask_names (FITSMaskNames): The previously masked image

    Returns:
        FITSMaskNames: _description_
    """
    # TODO: Ideally we can accept an arbitrary WCS, or read the wsclean docs to
    # try to construct it ourselves. The last thing that this pirate wants is
    # to run the imager in a dry-run type mode n cleaning type mode purely for
    # the WCS.

    mask_names = create_fits_mask_names(fits_image=fits_beam_image_path)

    with fits.open(fits_beam_image_path) as beam_image:
        header = beam_image[0].header
        beam_image_shape = beam_image[0].data.shape[-2:]

    logger.info(
        f"Will extract {beam_image_shape} from {fits_mosaic_mask_names.mask_fits}"
    )

    with fits.open(fits_mosaic_mask_names.mask_fits) as mosaic_mask:
        logger.info("Extracting region")
        extract_image = reproject_interp(
            input_data=(
                np.squeeze(mosaic_mask[0].data),
                WCS(mosaic_mask[0].header).celestial,
            ),
            output_projection=WCS(header).celestial,
            shape_out=beam_image_shape,
            order=0,
        )
    logger.info("Clipping extract mask results (interpolation errors)")
    extract_image = np.clip(extract_image, 0.0, 1.0)

    fits.writeto(mask_names.mask_fits, extract_image[0].astype(np.int32), header)

    return mask_names


def _get_signal_image(
    image: np.ndarray | None = None,
    rms: np.ndarray | None = None,
    background: np.ndarray | None = None,
    signal: np.ndarray | None = None,
) -> np.ndarray:
    if all([item is None for item in (image, background, rms, signal)]):
        raise ValueError("No input maps have been provided. ")

    if signal is None and image is not None and rms is not None:
        if background is None:
            logger.info("No background supplied, assuming zeros. ")
            background = np.zeros_like(image)

        out_signal = (image - background) / rms
    else:
        out_signal = signal

    return out_signal


def grow_low_snr_mask(
    image: np.ndarray | None = None,
    rms: np.ndarray | None = None,
    background: np.ndarray | None = None,
    signal: np.ndarray | None = None,
    grow_low_snr: float = 2.0,
    grow_low_island_size: int = 512,
    region_mask: np.ndarray | None = None,
) -> np.ndarray:
    """There may be cases where normal thresholding operations based on simple pixel-wise SNR
    cuts fail to pick up diffuse, low surface brightness regions of emission. When some type
    of masking operation is used there may be instances where these regions are never cleaned.

    Sometimes smoothing can help to pick up these features, but when attempting to pass such
    a mask through to the imagery of choice the islands may be larger than the features at their
    native resolution, unless some other more sophisticated filtering is performed.

    This function attempts to grow masks to capture islands of contiguous pixels above a low
    SNR cut that would otherwise go uncleaned.

    Args:
            Args:
        image (Optional[np.ndarray], optional): The total intensity pixels to have the mask for. Defaults to None.
        rms (Optional[np.ndarray], optional): The noise across the image. Defaults to None.
        background (Optional[np.ndarray], optional): The background acros the image. If None, zeros are assumed. Defaults to None.
        signal(Optional[np.ndarray], optional): A signal map. Defaults to None.
        grow_low_snr (float, optional): The SNR pixekls have to be above. Defaults to 2.
        grow_low_island_size (int, optional): The minimum number of pixels an island should be for it to be considered valid. Defaults to 512.
        region_mask (Optional[np.ndarray], optional): Whether some region should be masked out before the island size constraint is applied. Defaults to None.

    Returns:
        np.ndarray: The derived mask of objects with low-surface brightness
    """
    # TODO: The `grow_low_island_size` should be represented in solid angle relative to the restoring beam

    signal = _get_signal_image(
        image=image, rms=rms, background=background, signal=signal
    )

    logger.info(
        f"Growing mask for low surface brightness using {grow_low_snr=} {grow_low_island_size=}"
    )
    low_snr_mask = signal > grow_low_snr
    low_snr_mask = scipy_binary_dilation(
        input=low_snr_mask, iterations=2, structure=np.ones((3, 3))
    )
    low_snr_mask = scipy_binary_erosion(
        input=low_snr_mask, iterations=2, structure=np.ones((3, 3))
    )

    if region_mask is not None:
        low_snr_mask[region_mask] = False

    mask_labels, no_labels = label(low_snr_mask, structure=np.ones((3, 3)))
    _, counts = np.unique(mask_labels.flatten(), return_counts=True)

    small_islands = [
        idx
        for idx, count in enumerate(counts)
        if count < grow_low_island_size and idx > 0
    ]
    low_snr_mask[np.isin(mask_labels, small_islands)] = False

    return low_snr_mask


class SkewResult(NamedTuple):
    positive_pixel_frac: np.ndarray
    """The fraction of positive pixels in a boxcar function"""
    skew_mask: np.ndarray
    """Mask of pixel positions indicating which positions failed the skew test"""
    box_size: int
    """Size of the boxcar window applies"""
    skew_delta: float
    """The test threshold for skew"""


def create_boxcar_skew_mask(
    image: np.ndarray, skew_delta: float, box_size: int
) -> np.ndarray:
    assert 0.0 < skew_delta < 0.5, f"{skew_delta=}, but should be 0.0 to 0.5"
    assert len(image.shape) == 2, (
        f"Expected two dimensions, got image shape of {image.shape}"
    )
    logger.info(f"Computing boxcar skew with {box_size=} and {skew_delta=}")
    positive_pixels = (image > 0.0).astype(np.float32)

    # Counting positive pixel fraction here. The su
    window_shape = (box_size, box_size)
    positive_pixel_fraction = fftconvolve(
        in1=positive_pixels, in2=np.ones(window_shape, dtype=np.float32), mode="same"
    ) / np.prod(window_shape)
    positive_pixel_fraction = np.clip(
        positive_pixel_fraction, 0.0, 1.0
    )  # trust nothing

    skew_mask = positive_pixel_fraction > (0.5 + skew_delta)
    logger.info(f"{np.sum(skew_mask)} pixels above {skew_delta=} with {box_size=}")

    return SkewResult(
        positive_pixel_frac=positive_pixel_fraction,
        skew_mask=skew_mask,
        skew_delta=skew_delta,
        box_size=box_size,
    )


def _minimum_absolute_clip(
    image: np.ndarray, increase_factor: float = 2.0, box_size: int = 100
) -> np.ndarray:
    """Given an input image or signal array, construct a simple image mask by applying a
    rolling boxcar minimum filter, and then selecting pixels above a cut of
    the absolute value value scaled by `increase_factor`. This is a pixel-wise operation.

    Args:
        image (np.ndarray): The input array to consider
        increase_factor (float, optional): How large to scale the absolute minimum by. Defaults to 2.0.
        box_size (int, optional): Size of the rolling boxcar minimum filtr. Defaults to 100.

    Returns:
        np.ndarray: The mask of pixels above the locally varying threshold
    """
    logger.info(f"Minimum absolute clip, {increase_factor=} {box_size=}")
    rolling_box_min = minimum_filter(image, box_size)

    image_mask = image > (increase_factor * np.abs(rolling_box_min))
    # NOTE: This used to attempt to select pixels should that belong to an island of positive pixels with a box that was too small
    # | (
    #     (image > 0.0) & (rolling_box_min > 0.0)
    # )

    return image_mask


def _adaptive_minimum_absolute_clip(
    image: np.ndarray,
    increase_factor: float = 2.0,
    box_size: int = 100,
    adaptive_max_depth: int = 3,
    adaptive_box_step: float = 2.0,
    adaptive_skew_delta: float = 0.2,
) -> np.ndarray:
    logger.info(
        f"Using adaptive minimum absolute clip with {box_size=} {adaptive_skew_delta=}"
    )
    min_value = minimum_filter(input=image, size=box_size)

    for box_round in range(adaptive_max_depth, 0, -1):
        skew_results = create_boxcar_skew_mask(
            image=image, skew_delta=adaptive_skew_delta, box_size=box_size
        )
        if np.all(~skew_results.skew_mask):
            logger.info("No skewed islands detected")
            break
        if any([box_size > dim for dim in image.shape]):
            logger.info(f"{box_size=} larger than a dimension in {image.shape=}")
            break

        logger.info(f"({box_round}) Growing {box_size=} {adaptive_box_step=}")
        box_size = int(box_size * adaptive_box_step)
        _min_value = minimum_filter(input=image, size=box_size)
        logger.debug("Slicing minimum values into place")

        min_value[skew_results.skew_mask] = _min_value[skew_results.skew_mask]

    mask = image > (np.abs(min_value) * increase_factor)

    return mask


def minimum_absolute_clip(
    image: np.ndarray,
    increase_factor: float = 2.0,
    box_size: int = 100,
    adaptive_max_depth: int | None = None,
    adaptive_box_step: float = 2.0,
    adaptive_skew_delta: float = 0.2,
) -> np.ndarray:
    """Implements minimum absolute clip method. A minimum filter of a particular
    boxc size is applied to the input image. The absolute of the output is taken
    and increased by a guard factor, which forms the clipping level used to construct
    a clean mask:

    >>> image > (absolute(minimum_filter(image, box)) * factor)

    The idea is only valid for zero mean and normally distributed pixels, with
    positive definite flux, making it appropriate for Stokes I.

    Larger box sizes and guard factors will make the mask more conservative. Should
    the boxcar be too small relative to some feature it is aligned it is possible
    that an excess of positive pixels will produce an less than optimal clipping
    level. An adaptive box size mode, if activated, attempts to use a larger box
    around these regions.

    The basic idea being detecting regions where the boxcar is too small is around
    the idea that there should be a similar number of positive to negative pixels.
    Should there be too many positive pixels in a region it is likely there is an

    Args:
        image (np.ndarray): Image to create a mask for
        increase_factor (float, optional): The guard factor used to inflate the absolute of the minimum filter. Defaults to 2.0.
        box_size (int, optional): Size of the box car of the minimum filter. Defaults to 100.
        adaptive_max_depth (Optional[int], optional): The maximum number of rounds that the adaptive mode is allowed to perform when rescaling boxcar results in certain directions. Defaults to None.
        adaptive_box_step (float, optional): A multiplicative factor to increase the boxcar size by each round. Defaults to 2.0.
        adaptive_skew_delta (float, optional): Minimum deviation from 0.5 that needs to be met to classify a region as skewed. Defaults to 0.2.

    Returns:
        np.ndarray: Final mask
    """

    if adaptive_max_depth is None:
        return _minimum_absolute_clip(
            image=image, box_size=box_size, increase_factor=increase_factor
        )

    adaptive_max_depth = int(adaptive_max_depth)

    return _adaptive_minimum_absolute_clip(
        image=image,
        increase_factor=increase_factor,
        box_size=box_size,
        adaptive_max_depth=adaptive_max_depth,
        adaptive_box_step=adaptive_box_step,
        adaptive_skew_delta=adaptive_skew_delta,
    )


def _verify_set_positive_seed_clip(
    positive_seed_clip: float, signal: np.ndarray
) -> float:
    """Ensure that the positive seed clip is handled appropriately"""
    max_signal = np.max(signal)
    if max_signal < positive_seed_clip:
        logger.critical(
            f"The maximum signal {max_signal:.4f} is below the provided {positive_seed_clip=}. "
            "Setting clip to 90 percent of maximum. "
        )
        positive_seed_clip = max_signal * 0.9

    return positive_seed_clip


def reverse_negative_flood_fill(
    base_image: np.ndarray,
    masking_options: MaskingOptions,
    pixels_per_beam: float | None = None,
) -> np.ndarray:
    """Attempt to:

    * seed masks around bright regions of an image and grow them to lower significance thresholds
    * remove regions of negative and positive islands that surround bright sources.

    An initial set of islands (and masks) are constructed by first
    using the `positive_seed_clip` to create an initial SNR based
    mask. These islands then are binary dilated to grow the islands
    to adjacent pixels at a lower significance level (see `scipy.ndimage.binary_dilation`).

    Next an attempt is made to remove artefacts around bright sources,  where
    there are likely to be positive and negative artefacts
    that originate from calibration errors, deconvolution errors, or residual
    structure from an incomplete clean.

    This operation will search for islands of _negative_ pixels above a
    threshold. These pixels are then grown after a guard mask has been constructed
    around bright pixels.

    The assumptions that go into this process:

    * the no genuine source of negative sky emission
    * if there are bright negative artefacts there are likely bright positive artefacts nearby
    * such negative pixels are ~10% level artefacts from a presumed bright sources

    Optionally, the `grow_low_snr_mask` may also be considered via the `grow_low_snr` and `grow_low_island_size`
    parameters.

    Args:
        base_image (np.ndarray): The base image or signal map that is used throughout the fill procedure.
        masking_options (MaskingOptions): Options to carry out masking.
        pixels_per_beam (Optional[float], optional): The number of pixels that cover a beam. Defaults to None.

    Returns:
        np.ndarray: Mask of the pixels to clean
    """

    logger.info("Will be reversing flood filling")
    logger.info(f"{masking_options=} ")

    if masking_options.flood_fill_use_mbc:
        positive_mask = minimum_absolute_clip(
            image=base_image,
            increase_factor=masking_options.flood_fill_positive_seed_clip,
            box_size=masking_options.flood_fill_use_mbc_box_size,
            adaptive_max_depth=masking_options.flood_fill_use_mbc_adaptive_max_depth,
            adaptive_box_step=masking_options.flood_fill_use_mbc_adaptive_step_factor,
            adaptive_skew_delta=masking_options.flood_fill_use_mbc_adaptive_skew_delta,
        )
        flood_floor_mask = minimum_absolute_clip(
            image=base_image,
            increase_factor=masking_options.flood_fill_positive_flood_clip,
            box_size=masking_options.flood_fill_use_mbc_box_size,
            adaptive_max_depth=masking_options.flood_fill_use_mbc_adaptive_max_depth,
            adaptive_box_step=masking_options.flood_fill_use_mbc_adaptive_step_factor,
            adaptive_skew_delta=masking_options.flood_fill_use_mbc_adaptive_skew_delta,
        )
    else:
        # Sanity check the upper clip level, you rotten seadog
        positive_seed_clip = _verify_set_positive_seed_clip(
            positive_seed_clip=masking_options.flood_fill_positive_seed_clip,
            signal=base_image,
        )
        # Here we create the mask image that will start the binary dilation
        # process, and we will ensure only pixels above the `positive_flood_clip`
        # are allowed to be dilated. In other words we are growing the mask
        positive_mask = base_image >= positive_seed_clip
        flood_floor_mask = base_image > masking_options.flood_fill_positive_flood_clip

    positive_dilated_mask = scipy_binary_dilation(
        input=positive_mask,
        mask=flood_floor_mask,
        iterations=1000,
        structure=np.ones((3, 3)),
    )

    if masking_options.grow_low_snr_island:
        low_snr_mask = grow_low_snr_mask(
            signal=base_image,
            grow_low_snr=masking_options.grow_low_snr_island_clip,
            grow_low_island_size=masking_options.grow_low_snr_island_size,
        )
        positive_dilated_mask[low_snr_mask] = True

    return positive_dilated_mask.astype(np.int32)


def _create_signal_from_rmsbkg(
    image: Path | np.ndarray,
    rms: Path | np.ndarray,
    bkg: Path | np.ndarray,
) -> np.ndarray:
    logger.info("Creating signal image")

    if isinstance(image, Path):
        with fits.open(image) as in_fits:
            logger.info(f"Loading {image}")
            image = in_fits[0].data  # type: ignore

    assert isinstance(image, np.ndarray), (
        f"Expected the image to be a numpy array by now, instead have {type(image)}"
    )

    if isinstance(bkg, Path):
        with fits.open(bkg) as in_fits:
            logger.info(f"Loading {bkg=}")
            bkg = in_fits[0].data  # type: ignore

    logger.info("Subtracting background")
    image -= bkg

    if isinstance(rms, Path):
        with fits.open(rms) as in_fits:
            logger.info(f"Loading {rms=}")
            rms = in_fits[0].data  # type: ignore

    logger.info("Dividing by rms")
    image /= rms

    return np.array(image)


def _need_to_make_signal(masking_options: MaskingOptions) -> bool:
    """Isolated functions to consider whether a signal image is needed"""
    return not masking_options.flood_fill_use_mbc


def create_snr_mask_from_fits(
    fits_image_path: Path,
    masking_options: MaskingOptions,
    fits_rms_path: Path | None,
    fits_bkg_path: Path | None,
    create_signal_fits: bool = False,
    overwrite: bool = True,
) -> FITSMaskNames:
    """Create a mask for an input FITS image based on a signal to noise given a corresponding pair of RMS and background FITS images.

    Internally should a signal image be needed it is computed as something akin to:
    > signal = (image - background) / rms

    This is done in a staged manner to minimise the number of (potentially large) images
    held in memory.

    Each of the input images needs to share the same shape. This means that compression
    features offered by some tooling (e.g. BANE --compress) can not be used.

    Depending on the `MaksingOptions` used the signal image may not be needed.

    Once the signal map as been computed, all pixels below ``min_snr`` are flagged.

    Args:
        fits_image_path (Path): Path to the FITS file containing an image
        masking_options (MaskingOptions): Configurables on the masking operation procedure.
        fits_rms_path (Optional[Path], optional): Path to the FITS file with an RMS image corresponding to ``fits_image_path``. Defaults to None.
        fits_bkg_path (Optional[Path], optional): Path to the FITS file with an background image corresponding to ``fits_image_path``. Defaults to None.
        create_signal_fits (bool, optional): Create an output signal map. Defaults to False.
        overwrite (bool): Passed to `fits.writeto`, and will overwrite files should they exist. Defaults to True.

    Returns:
        FITSMaskNames: Container describing the signal and mask FITS image paths. If ``create_signal_path`` is None, then the ``signal_fits`` attribute will be None.
    """
    mask_names = create_fits_mask_names(
        fits_image=fits_image_path, include_signal_path=create_signal_fits
    )

    # TODOL Make the bkg and rms images optional. Don't need to load if mbc is usede
    with fits.open(fits_image_path) as fits_image:
        fits_header = fits_image[0].header  # type: ignore
        signal_data = fits_image[0].data  # type: ignore

    if _need_to_make_signal(masking_options=masking_options):
        assert isinstance(fits_rms_path, Path) and isinstance(fits_bkg_path, Path), (
            "Expected paths for input RMS and bkg FITS files"
        )
        signal_data = _create_signal_from_rmsbkg(
            image=signal_data, rms=fits_rms_path, bkg=fits_bkg_path
        )

        if create_signal_fits:
            logger.info(f"Writing {mask_names.signal_fits}")
            fits.writeto(
                filename=mask_names.signal_fits,
                data=signal_data,
                header=fits_header,
                overwrite=overwrite,
            )

    pixels_per_beam = get_pixels_per_beam(fits_path=fits_image_path)

    # Following the help in wsclean:
    # WSClean accepts masks in CASA format and in fits file format. A mask is a
    # normal, single polarization image file, where all zero values are interpreted
    # as being not masked, and all non-zero values are interpreted as masked. In the
    # case of a fits file, the file may either contain a single frequency or it may
    # contain a cube of images.
    if masking_options.flood_fill:
        # TODO: The image and signal masks both don't need to be inputs. Image is only used
        # if mbc = True
        mask_data = reverse_negative_flood_fill(
            base_image=np.squeeze(signal_data),
            masking_options=masking_options,
            pixels_per_beam=pixels_per_beam,
        )
        mask_data = mask_data.reshape(signal_data.shape)
    else:
        logger.info(f"Clipping using a {masking_options.base_snr_clip=}")
        mask_data = (signal_data > masking_options.base_snr_clip).astype(np.int32)

    if masking_options.beam_shape_erode:
        mask_data = beam_shape_erode(
            mask=mask_data,
            fits_header=fits_header,
            minimum_response=masking_options.beam_shape_erode_minimum_response,
        )

    logger.info(f"Writing {mask_names.mask_fits}")
    fits.writeto(
        filename=mask_names.mask_fits,
        data=mask_data,
        header=fits_header,
        overwrite=overwrite,
    )

    return mask_names


def get_parser() -> ArgumentParser:
    parser = ArgumentParser(
        description="Simple utility functions to create masks from FITS images. "
    )

    subparser = parser.add_subparsers(
        dest="mode", help="Operation mode of flint_bandpass"
    )

    fits_parser = subparser.add_parser(
        "mask",
        help="Create a mask for an image, potentially using its RMS and BKG images (e.g. outputs from BANE). Output FITS image will default to the image with a mask suffix.",
    )
    fits_parser.add_argument("image", type=Path, help="Path to the input image. ")
    fits_parser = add_options_to_parser(
        parser=fits_parser, options_class=MaskingOptions
    )

    fits_parser.add_argument(
        "--rms-fits", type=Path, help="Path to the RMS of the input image. "
    )
    fits_parser.add_argument(
        "--bkg-fits", type=Path, help="Path to the BKG of the input image. "
    )
    fits_parser.add_argument(
        "--save-signal",
        action="store_true",
        help="Save the signal image internally generated (should it be generated)",
    )

    extract_parser = subparser.add_parser(
        "extractmask",
        help="Extract a beam FITS masked region from a larger FITS mask mosaic image. ",
    )

    extract_parser.add_argument(
        "beam_image",
        type=Path,
        help="The FITS image with the WCS that will be used to extract the mask region from",
    )

    extract_parser.add_argument(
        "mosaic_image",
        type=Path,
        help="The field mosaic image that will be used as a basis to extract a region mask from",
    )

    return parser


def cli():
    parser = get_parser()

    args = parser.parse_args()

    if args.mode == "mask":
        masking_options = create_options_from_parser(
            parser_namespace=args, options_class=MaskingOptions
        )
        create_snr_mask_from_fits(
            fits_image_path=args.image,
            fits_rms_path=args.rms_fits,
            fits_bkg_path=args.bkg_fits,
            create_signal_fits=args.save_signal,
            masking_options=masking_options,
        )

    elif args.mode == "extractmask":
        extract_beam_mask_from_mosaic(
            fits_beam_image_path=args.beam_image,
            fits_mosaic_mask_names=FITSMaskNames(mask_fits=args.mosaic_image),
        )
    else:
        logger.error(f"Supplied mode {args.mode} is not known. ")


if __name__ == "__main__":
    cli()



"""Test utilities related to flagging measurement set operations"""

from __future__ import annotations

import shutil
from pathlib import Path

import numpy as np
import pytest
from casacore.tables import table

from flint.containers import get_known_container_path
from flint.flagging import (
    flag_ms_aoflagger,
    flag_ms_zero_uvws,
    nan_zero_extreme_flag_ms,
)
from flint.ms import MS
from flint.utils import get_packaged_resource_path

from .test_containers import flint_containers
from .test_helpers import which

_ = flint_containers  # make sure fixture is imported and not linted away


@pytest.fixture
def ms_example(tmpdir):
    ms_zip = Path(
        get_packaged_resource_path(
            package="flint.data.tests",
            filename="SB39400.RACS_0635-31.beam0.small.ms.zip",
        )
    )
    outpath = Path(tmpdir) / "39400"

    shutil.unpack_archive(ms_zip, outpath)

    ms_path = Path(outpath) / "SB39400.RACS_0635-31.beam0.small.ms"

    return ms_path


def test_flag_ms_zero_uvws(ms_example):
    flag_ms_zero_uvws(ms=ms_example)

    with table(str(ms_example), ack=False) as tab:
        uvws = tab.getcol("UVW")
        flags = tab.getcol("FLAG")

        uvw_mask = np.all(uvws == 0, axis=1)

        assert np.all(flags[uvw_mask] == True)  # noQA: E712


def test_add_flag_ms_zero_uvws(ms_example):
    with table(str(ms_example), readonly=False, ack=False) as tab:
        uvws = tab.getcol("UVW")
        uvws[:1, :] = 0

        tab.putcol("UVW", uvws)

    flag_ms_zero_uvws(ms=ms_example)

    with table(str(ms_example), ack=False) as tab:
        uvws = tab.getcol("UVW")

        flags = tab.getcol("FLAG")

        uvw_mask = np.all(uvws == 0, axis=1)

        assert np.sum(uvw_mask) > 0
        assert np.all(flags[uvw_mask] == True)  # noQA: E712


def test_nan_zero_extreme_flag_ms(ms_example):
    """Makes sure that flagging the NaNs, UVWs of zero and
    extreme outliers works. Was added after introducing the
    chunking"""
    with table(str(ms_example), readonly=False, ack=False) as tab:
        uvws = tab.getcol("UVW")
        uvws[:10, :] = 0

        tab.putcol("UVW", uvws)

    nan_zero_extreme_flag_ms(ms=ms_example)

    with table(str(ms_example), ack=False) as tab:
        uvws = tab.getcol("UVW")

        flags = tab.getcol("FLAG")

        uvw_mask = np.all(uvws == 0, axis=1)

        assert np.sum(uvw_mask) > 0
        assert np.all(flags[uvw_mask] == True)  # noQA: E712


def test_nan_zero_extreme_flag_ms_with_chunks(ms_example):
    """Makes sure that flagging the NaNs, UVWs of zero and
    extreme outliers works. Was added after introducing the
    chunking.

    Same as above test but with chunk size"""
    with table(str(ms_example), readonly=False, ack=False) as tab:
        uvws = tab.getcol("UVW")
        uvws[:20, :] = 0

        tab.putcol("UVW", uvws)

    nan_zero_extreme_flag_ms(ms=ms_example, chunk_size=1)

    with table(str(ms_example), ack=False) as tab:
        uvws = tab.getcol("UVW")

        flags = tab.getcol("FLAG")

        uvw_mask = np.all(uvws == 0, axis=1)

        assert np.sum(uvw_mask) > 0
        assert np.all(flags[uvw_mask] == True)  # noQA: E712


def test_nan_zero_extreme_flag_ms_with_chunks_and_datanan(ms_example):
    """Makes sure that flagging the NaNs, UVWs of zero and
    extreme outliers works. Was added after introducing the
    chunking.

    Same as above test but with chunk size and naning data when flags are true"""
    with table(str(ms_example), readonly=False, ack=False) as tab:
        uvws = tab.getcol("UVW")
        uvws[:20, :] = 0
        original_data = tab.getcol("DATA")
        tab.putcol("UVW", uvws)

    nan_zero_extreme_flag_ms(ms=ms_example, chunk_size=1, nan_data_on_flag=True)

    with table(str(ms_example), ack=False) as tab:
        uvws = tab.getcol("UVW")

        flags = tab.getcol("FLAG")
        data = tab.getcol("DATA")
        uvw_mask = np.all(uvws == 0, axis=1)

        assert np.sum(np.isnan(data)) == np.sum(flags)
        assert np.sum(np.isnan(data)) > np.sum(np.isnan(original_data))
        assert np.sum(uvw_mask) > 0
        assert np.all(flags[uvw_mask] == True)  # noQA: E712


# Vontainer related options
if which("singularity") is None:
    pytest.skip("Singularity is not installed", allow_module_level=True)


def test_aoflagger(flint_containers, ms_example) -> None:
    """Ensure we can run a aoflagger recipe"""

    aoflagger_path = get_known_container_path(
        container_directory=flint_containers, name="aoflagger"
    )
    assert isinstance(aoflagger_path, Path)
    assert aoflagger_path.exists()

    ms = MS(path=ms_example, column="DATA")

    with table(str(ms.path), readonly=False) as tab:
        orig_flags = tab.getcol("FLAG")
        tab.putcol("FLAG", np.zeros_like(orig_flags))
        # Reset the flags to ensure aoflagger does the
        # right set of things
        pre_flag = np.sum(orig_flags)

    flag_ms = flag_ms_aoflagger(ms=ms, container=aoflagger_path)

    assert isinstance(flag_ms, MS)
    assert flag_ms.path == ms.path

    with table(str(ms.path), readonly=True) as tab:
        # uvws shape is (row, coord)
        uvws = tab.getcol("UVW")
        # flags shapoe is {row, channel, pol}
        flags = tab.getcol("FLAG")
        post_flag = np.sum(flags)
        u = uvws[:, 0]
        assert np.all(u[~flags[:, 0, 0]] != 0)

    assert pre_flag == post_flag



"""Simple tests for some prefect cluster helper functions"""

from __future__ import annotations

from pathlib import Path

import pytest
import yaml
from prefect_dask import DaskTaskRunner

from flint.prefect.clusters import get_cluster_spec, get_dask_runner


def test_example_yaml_from_file(tmpdir):
    """Place a Yaml file in a directory, and load it"""

    yaml_dir = Path(tmpdir) / "some_file1"
    yaml_dir.mkdir(parents=True, exist_ok=True)

    yaml_cluster = yaml_dir / "test.yaml"
    dump = dict(pirate="jack")

    with open(yaml_cluster, "w") as out_file:
        yaml.dump(dump, out_file)

    in_data = get_cluster_spec(cluster=yaml_cluster)

    assert "pirate" in in_data
    assert in_data["pirate"] == "jack"


def test_get_dask_task_runner():
    """Make sure we can fire up a dask task runner. This uses an example
    slurm cluster"""
    dask_task_runner = get_dask_runner(cluster="example_slurm")
    assert isinstance(dask_task_runner, DaskTaskRunner)

    dask_task_runner = get_dask_runner(
        cluster="example_slurm", extra_cluster_kwargs={"name": "jack-be-testing"}
    )
    assert isinstance(dask_task_runner, DaskTaskRunner)


def test_example_cluster_condif(tmpdir):
    """Load in a passable cluster configuration."""
    data = get_cluster_spec(cluster="example_slurm")
    assert data["cluster_class"] == "dask_jobqueue.SLURMCluster"


def test_example_cluster_no_exists():
    """Similar to the above but make sure an error is raised"""

    with pytest.raises(ValueError):
        get_cluster_spec(cluster="Jack-be-sneaky")



"""Tooling related to the convolution of images. Principally
this is mostly to smooth to a common resolution
"""

from __future__ import annotations

import warnings
from argparse import ArgumentParser
from pathlib import Path
from shutil import copyfile
from typing import Collection, Literal, NamedTuple

import astropy.units as u
import numpy as np
from astropy.io import fits
from astropy.wcs import FITSFixedWarning
from racs_tools import beamcon_2D, beamcon_3D
from radio_beam import Beam, Beams

from flint.logging import logger

warnings.simplefilter("ignore", FITSFixedWarning)


class BeamShape(NamedTuple):
    """A simple container to represent a fitted 2D gaussian,
    intended for the main lobe of the synthesised beam. This
    class has been defined to avoid issues with the serialisation
    of astropy.units, which can cause strange and wonderful
    errors when being sent over the wire to workers."""

    bmaj_arcsec: float
    """The size of the major-axis of the beam, in arcseconds."""
    bmin_arcsec: float
    """The size of the minor-axis of the beam, in arcseconds."""
    bpa_deg: float
    """Rotation of the beam, in degrees."""

    @classmethod
    def from_radio_beam(cls, radio_beam: Beam) -> BeamShape:
        """A helper function to convert a radio_beam.Beam into a
        BeamShape. This is prinicpally intended to be used when
        there is a need to exchange a Beam between processes
        that would need to serialise the object.

        Args:
            radio_beam (Beam): The Beam to convert to normalised and known units

        Returns:
            BeamShape: The normalised container without astropy units.
        """
        return cls(
            bmaj_arcsec=radio_beam.major.to(u.arcsecond).value,  # type: ignore
            bmin_arcsec=radio_beam.minor.to(u.arcsecond).value,  # type: ignore
            bpa_deg=radio_beam.pa.to(u.degree).value,  # type: ignore
        )


def check_if_cube_fits(fits_file: Path) -> bool:
    """Check to see whether the data component of a FITS images is a cube.
    Returns ``True`` is the data-shape needs 3-dimensions to be represented.

    Note: Unclear on usefulness

    Args:
        fits_file (Path): FITS file that will be examinined

    Returns:
        bool: Whether the input FITS file is a cube or not.
    """

    try:
        squeeze_data = np.squeeze(fits.getdata(fits_file))  # type: ignore
    except FileNotFoundError:
        return False
    except OSError:
        return False

    return len(squeeze_data.shape) == 3


def get_cube_common_beam(
    cube_paths: Collection[Path], cutoff: float | None = None
) -> list[BeamShape]:
    """Given a set of input cube FITS files, compute a common beam
    for each channel.

    Args:
        cube_paths (Collection[Path]): Set of cube FITS files to inspect to derive a common beam
        cutoff (Optional[float], optional): A cutoff value, in arcsec, that specifies the maximum BMAJ allowed. Defaults to None.

    Returns:
        List[BeamShape]: List of target beam shapes to use, corresponding to each channel
    """

    _, common_beam_data_list = beamcon_3D.smooth_fits_cube(
        infiles_list=list(cube_paths),
        dryrun=True,
        cutoff=cutoff,
        mode="natural",
        conv_mode="robust",
        ncores=1,
    )
    # Make proper check here that accounts for NaNs
    for file in common_beam_data_list:
        assert all(
            (file[0].major == common_beam_data_list[0][0].major)
            | np.isnan(file[0].major)
        )
        assert all(
            (file[0].minor == common_beam_data_list[0][0].minor)
            | np.isnan(file[0].minor)
        )
        assert all(
            (file[0].pa == common_beam_data_list[0][0].pa) | np.isnan(file[0].pa)
        )

    first_cube_fits_beam = common_beam_data_list[0][0]
    assert isinstance(first_cube_fits_beam, Beams), (
        f"Unexpected type for common beams. Expected Beams, got {type(first_cube_fits_beam)}"
    )

    beam_shape_list = [
        BeamShape.from_radio_beam(radio_beam=beam)  # type: ignore
        for beam in first_cube_fits_beam
    ]
    return beam_shape_list


def convolve_cubes(
    cube_paths: Collection[Path],
    beam_shapes: list[BeamShape],
    cutoff: float | None = None,
    convol_suffix: str = "conv",
    executor_type: Literal["thread", "process", "mpi"] = "thread",
) -> Collection[Path]:
    logger.info(f"Will attempt to convol {len(cube_paths)} cubes")
    if cutoff:
        logger.info(f"Supplied cutoff {cutoff}")

    # Extractubg the beam properties
    beam_major_list = [float(beam.bmaj_arcsec) for beam in beam_shapes]
    beam_minor_list = [float(beam.bmin_arcsec) for beam in beam_shapes]
    beam_pa_list = [float(beam.bpa_deg) for beam in beam_shapes]

    # Sanity test
    assert len(beam_major_list) == len(beam_minor_list) == len(beam_pa_list)

    logger.info("Convoling cubes")
    cube_data_list, _, _ = beamcon_3D.smooth_fits_cube(
        infiles_list=list(cube_paths),
        dryrun=False,
        cutoff=cutoff,
        mode="natural",
        conv_mode="robust",
        bmaj=beam_major_list,
        bmin=beam_minor_list,
        bpa=beam_pa_list,
        suffix=convol_suffix,
        executor_type=executor_type,
    )

    # Construct the name of the new file created. For the moment this is done
    # manually as it is not part of the returned object
    # TODO: Extend the return struct from beamcon_3D to include output name
    convol_cubes_path = [
        Path(cube_data.filename).with_suffix(f".{convol_suffix}.fits")
        for cube_data in cube_data_list
    ]

    # Show the mapping as a sanity check
    for input_cube, output_cube in zip(list(cube_paths), convol_cubes_path):
        logger.info(f"{input_cube=} convolved to {output_cube}")

    # Trust no one
    assert all([p.exists() for p in convol_cubes_path]), (
        "A convolved cube does not exist"
    )
    return convol_cubes_path


def get_common_beam(
    image_paths: Collection[Path], cutoff: float | None = None
) -> BeamShape:
    """Return the minimum beam size required to encompass the beams described
    in the FITS header (e.g. BMAJ,BMIN,BPA) of the input images. This is used
    when preparing to convolve to a common resolution.

    Args:
        image_paths (Collection[Path]): The paths to the FITS images that will be examined.
        cutoff (Optional[float], optional): The maximum beam size an input image is allowed to have. BMAJ's larger than this are ignored from the calculation. Defaults to None.

    Returns:
        BeamShape: Smalled common beam available to be used
    """

    logger.info(f"Calculating common beam size of {len(image_paths)} images. ")

    if cutoff:
        logger.info(f"Setting beam cutoff to {cutoff} arcseconds. ")

    try:
        beam, beams = beamcon_2D.get_common_beam(files=list(image_paths), cutoff=cutoff)

        beam_shape = BeamShape.from_radio_beam(beam)
        logger.info(f"Constructed {beam_shape=}")
    except ValueError:
        logger.info("The beam was not constrained. Setting to NaNs")
        beam_shape = BeamShape(bmaj_arcsec=np.nan, bmin_arcsec=np.nan, bpa_deg=np.nan)

    return beam_shape


def convolve_images(
    image_paths: Collection[Path],
    beam_shape: BeamShape,
    cutoff: float | None = None,
    convol_suffix: str = "conv",
    output_paths: list[Path] | None = None,
) -> list[Path]:
    """Convolve a set of input images to a common resolution as specified
    by the beam_shape. If the major-axis of the native resolution is larger
    than cutoff (in arcseconds) then the racs_tools beamconv_2D task will
    nan it.

    Additionally, some input subject image will simply copied if:

    * the input ``beam_shape`` is not finite, or
    * the beamshape encoded in the FITS header of the subject image is not defined

    Args:
        image_paths (Collection[Path]): Set of image paths to FITS images to convol
        beam_shape (BeamShape): The specification of the desired final resolution
        cutoff (Optional[float], optional): Images whose major-axis is larger than this will be blank. Expected in arcseconds. Defaults to None.
        convol_suffix (str, optional): The suffix added to .fits to indicate smoothed image. Defaults to 'conv'.
        output_paths (list[Path] | None, optional): The final output file namesfor each input image. If provided this renamed files created using the `convol_suffix`. Defaults to None.

    Returns:
        Collection[Path]: Set of paths to the smoothed images
    """

    logger.info(f"Will attempt to convolve {len(image_paths)} images.")
    if cutoff:
        logger.info(f"Supplied cutoff of {cutoff} arcsecond")

    if not np.isfinite(beam_shape.bmaj_arcsec):
        logger.info("Beam shape is not defined. Copying files into place. ")

        conv_image_paths = [
            Path(str(image_path).replace(".fits", f".{convol_suffix}.fits"))
            for image_path in image_paths
        ]
        # If the beam is not defined, simply copy the file into place. Although
        # this takes up more space, it is not more than otherwise
        for original_path, copy_path in zip(image_paths, conv_image_paths):
            logger.info(f"Copying {original_path=} {copy_path=}")
            copyfile(original_path, copy_path)

        return conv_image_paths

    radio_beam = Beam(
        major=beam_shape.bmaj_arcsec * u.arcsecond,
        minor=beam_shape.bmin_arcsec * u.arcsecond,
        pa=beam_shape.bpa_deg * u.deg,
    )

    return_conv_image_paths: list[Path] = []

    if output_paths:
        assert isinstance(output_paths, type(image_paths)), (
            "Types for image_paths and output_paths need to be the same"
        )
        assert len(output_paths) == len(image_paths), (
            f"Mismatch collection lengths of image_paths ({len(image_paths)}) and output_paths ({len(output_paths)})"
        )

    for idx, image_path in enumerate(image_paths):
        convol_output_path: Path = Path(
            str(image_path).replace(".fits", f".{convol_suffix}.fits")
        )
        header = fits.getheader(image_path)
        if header["BMAJ"] == 0.0:
            logger.info(f"Copying {image_path} to {convol_output_path=} for empty beam")
            copyfile(image_path, convol_output_path)
        else:
            logger.info(f"Convolving {image_path.name!s}")
            beamcon_2D.beamcon_2d_on_fits(
                file=image_path,
                outdir=None,
                new_beam=radio_beam,
                conv_mode="robust",
                suffix=convol_suffix,
                cutoff=cutoff,
            )

        if output_paths:
            output_path: Path = output_paths[idx]
            logger.info(f"Renaming generate convolved file to {output_path=}")
            convol_output_path.rename(output_path)
            convol_output_path = output_path

            # Pirates trust nothing, especially with the silly logic
            assert convol_output_path.exists(), (
                f"{convol_output_path=} should exist, but doesn't"
            )

        return_conv_image_paths.append(convol_output_path)

    return return_conv_image_paths


def get_parser() -> ArgumentParser:
    parser = ArgumentParser(description=__doc__)

    subparsers = parser.add_subparsers(dest="mode")

    convol_parser = subparsers.add_parser(
        "convol", help="Convol images to a common resolution"
    )

    convol_parser.add_argument(
        "images",
        type=Path,
        nargs="+",
        help="The images that will be convolved to a common resolution",
    )
    convol_parser.add_argument(
        "--cutoff",
        type=float,
        default=None,
        help="Beams whose major-axis are larger then this (in arcseconds) are ignored from the calculation of the optimal beam.",
    )
    convol_parser.add_argument(
        "--convol-suffix",
        type=str,
        default="conv",
        help="The suffix added to convolved images. ",
    )
    convol_parser.add_argument(
        "--cubes",
        action="store_true",
        default=False,
        help="Treat the input files as cubes and use the corresponding 3D beam selection and convolution. ",
    )

    maxbeam_parser = subparsers.add_parser(
        "maxbeam", help="Find the optimal beam size for a set of images."
    )

    maxbeam_parser.add_argument(
        "images",
        type=Path,
        nargs="+",
        help="The images that will be convolved to a common resolution",
    )
    maxbeam_parser.add_argument(
        "--cutoff",
        type=float,
        default=None,
        help="Beams whose major-axis are larger then this (in arcseconds) are ignored from the calculation of the optimal beam.",
    )

    cubemaxbeams_parser = subparsers.add_parser(
        "cubemaxbeam",
        help="Calculate the set of common beams across channels in a set of cubes",
    )
    cubemaxbeams_parser.add_argument(
        "cubes",
        type=Path,
        nargs="+",
        help="The images that will be convolved to a common resolution",
    )
    cubemaxbeams_parser.add_argument(
        "--cutoff",
        type=float,
        default=None,
        help="Beams whose major-axis are larger then this (in arcseconds) are ignored from the calculation of the optimal beam.",
    )

    return parser


def cli() -> None:
    parser = get_parser()

    args = parser.parse_args()

    if args.mode == "maxbeam":
        get_common_beam(image_paths=args.images, cutoff=args.cutoff)
    if args.mode == "convol":
        if args.cubes:
            assert all([check_if_cube_fits(fits_file=f) for f in args.images]), (
                "Not all input files are FITS cubes"
            )
            common_beams = get_cube_common_beam(
                cube_paths=args.images, cutoff=args.cutoff
            )
            for image in args.images:
                logger.info(f"Convoling {image}")
                _ = convolve_cubes(
                    cube_paths=[image],
                    beam_shapes=common_beams,
                    cutoff=args.cutoff,
                    convol_suffix=args.convol_suffix,
                )

        else:
            assert not all([check_if_cube_fits(fits_file=f) for f in args.images]), (
                "Not all input files are FITS images (not cubes)"
            )
            common_beam = get_common_beam(image_paths=args.images, cutoff=args.cutoff)
            _ = convolve_images(
                image_paths=args.images,
                beam_shape=common_beam,
                cutoff=args.cutoff,
                convol_suffix=args.convol_suffix,
            )
    if args.mode == "cubemaxbeam":
        common_beam_shape_list = get_cube_common_beam(
            cube_paths=args.cubes, cutoff=args.cutoff
        )
        logger.info(f"Extracted {common_beam_shape_list=}")


if __name__ == "__main__":
    cli()



"""Summary containers intended to hold general information obtained throughout a processing run."""

from __future__ import (  # Used for mypy/pylance to like the return type of MS.with_options
    annotations,
)

from pathlib import Path
from typing import NamedTuple

import astropy.units as u
from astropy.coordinates import (
    AltAz,
    EarthLocation,
    Latitude,
    Longitude,
    SkyCoord,
    concatenate,
)
from astropy.table import Table
from astropy.time import Time

# Addressing some time interval IERS issue with astropy.
from astropy.utils.iers import conf

from flint.coadd.linmos import LinmosResult
from flint.imager.wsclean import ImageSet, WSCleanResult
from flint.logging import logger
from flint.ms import (
    MS,
    MSSummary,
    describe_ms,
    get_pol_axis_from_ms,
    get_telescope_location_from_ms,
    get_times_from_ms,
)
from flint.naming import get_sbid_from_path, processed_ms_format
from flint.source_finding.aegean import AegeanOutputs
from flint.utils import estimate_skycoord_centre

conf.auto_max_age = None


class FieldSummary(NamedTuple):
    """The main information about a processed field. This structure is
    intended to store critical components that might be accumulated throughout
    processing of a pipeline, and may be most useful when data-products
    are zipped (or removed) throughout. Its intended usage is to hold key
    components that might be used in stages like validation plotting. It is not
    intended to become a catch-all to replace passing through items into
    functions directly.

    There are known issues around serialising astropy units within a dask/prefect environment,
    for example: https://github.com/astropy/astropy/issues/11317,
    This could become important if an instance of this object is exchanged
    between many prefect or dask like delayed tasks. Ye be warned.
    """

    sbid: str
    """SBID of processed field"""
    cal_sbid: str
    """SBID of the bandpass calibrator"""
    field_name: str
    """The name of the field"""
    ms_summaries: tuple[MSSummary, ...] | None = None
    """Summaries of measurement sets used in the processing of the filed"""
    centre: SkyCoord | None = None
    """Centre of the field, which is calculated as the mean position of all phase directions of the `mss` measurement sets"""
    integration_time: int | None = None
    """The integration time of the observation (seconds)"""
    no_components: int | None = None
    """Number of components found from the source finder"""
    holography_path: Path | None = None
    """Path to the file used for holography"""
    round: int | None = None
    """The self-cal round"""
    location: EarthLocation | None = None
    """The location of the telescope stored as (X,Y,Z) in meters"""
    ms_times: Time | None = None
    """The unique scan times of integrations stored in the measurement set"""
    hour_angles: Longitude | None = None
    """Computed hour-angles of the field"""
    elevations: Latitude | None = None
    """Computed elevations of the field"""
    median_rms: float | None = None
    """The meanian RMS computed from an RMS image"""
    beam_summaries: list[BeamSummary] | None = None
    """Summary information from each beam. Contains MSSummary, ImageSet and other information."""
    linmos_image: Path | None = None
    """The path to the linmos image of all beams"""
    pol_axis: float | None = None
    """The orientation of the ASKAP third-axis in radians. """

    def with_options(self, **kwargs) -> FieldSummary:
        prop = self._asdict()
        prop.update(**kwargs)

        return FieldSummary(**prop)


def _get_pol_axis_as_rad(ms: MS | Path) -> float:
    """Helper to get the appropriate pol_axis out of a MS. Prioritises the instrumental third-axis imprinted from fixms"""
    ms = MS.cast(ms=ms)

    # The INSTRUMENT_RECEPTOR_ANGLE comes from fixms and is
    # inserted to preserve the original orientation.

    try:
        pol_axis = get_pol_axis_from_ms(ms=ms, col="INSTRUMENT_RECEPTOR_ANGLE")

        logger.info(f"INSTRUMENT_RECEPTOR_ANGLE obtained pol_axis={pol_axis}")
    except ValueError:
        pol_axis = get_pol_axis_from_ms(ms=ms, col="RECEPTOR_ANGLE")

        # The prefect logger (or maybe the logger in general) does not render the quantity
        logger.info(f"RECEPTOR_ANGLE obtained pol_axis={pol_axis}")

    return pol_axis.to(u.rad).value


# TODO: Need to establise a MSLike type
def add_ms_summaries(field_summary: FieldSummary, mss: list[MS]) -> FieldSummary:
    """Obtain a MSSummary instance to add to a FieldSummary

    Quantities derived from the field centre (hour angles, elevations) are
    also calculated. The field centre position is estimated by taking the
    mean position of all phase directions.

    See `flint.utils.estimate_skycoord_centre`

    Args:
        field_summary (FieldSummary): Existing field summary object to update
        mss (List[MS]): Set of measurement sets to describe

    Returns:
        Tuple[MSSummary]: Results from the inspected set of measurement sets
    """
    logger.info("Adding MS summaries")

    ms_summaries = tuple(map(describe_ms, mss))
    centres_list = [ms_summary.phase_dir for ms_summary in ms_summaries]
    if len(centres_list) == 0:
        raise ValueError("No phase directions found in the MSs")
    elif len(centres_list) == 1:
        centre = centres_list[0]
    else:
        centres = concatenate(centres_list)
        centre = estimate_skycoord_centre(sky_positions=centres)

    telescope = field_summary.location
    ms_times = field_summary.ms_times
    centre_altaz = centre.transform_to(AltAz(obstime=ms_times, location=telescope))
    hour_angles = centre_altaz.az.to(u.hourangle)  # type: ignore
    elevations = centre_altaz.alt.to(u.deg)  # type: ignore

    # The INSTRUMENT_RECEPTOR_ANGLE comes from fixms and is
    # inserted to preserve the original orientation.
    pol_axis = _get_pol_axis_as_rad(ms=mss[0])

    field_summary = field_summary.with_options(
        ms_summaries=ms_summaries,
        centre=centre,
        hour_angles=hour_angles,
        elevations=elevations,
        pol_axis=pol_axis,
    )

    return field_summary


def add_rms_information(
    field_summary: FieldSummary, aegean_outputs: AegeanOutputs
) -> FieldSummary:
    """Add information derived from an RMS image and component catalogue
    to an existing `FieldSummary` instance. Some properteries, such as
    the center position, number of components etc, are taken directly
    from source finder products.

    On the center position -- there is not (at the moment) a simple
    way of getting the center position of a field. So the image
    itself is used to grab it.

    Other properties that require components of a measurement set,
    including the time and position, are also derived using existing
    fields which are created when a new instance is made.

    Args:
        field_summary (FieldSummary): Existing field summary object to update
        aegean_outputs (AegeanOutputs): Products of a source finding run

    Returns:
        FieldSummary: Updated field summary object
    """
    no_components = len(Table.read(aegean_outputs.comp))

    field_summary = field_summary.with_options(
        no_components=no_components,
    )

    return field_summary


def add_linmos_fits_image(
    field_summary: FieldSummary, linmos_command: LinmosResult
) -> FieldSummary:
    """Extract the path of the linmos fits image from the LinmosResult
    the co-added the field

    Args:
        field_summary (FieldSummary): Existing field summary to update
        linmos_command (LinmosResult): Instance of a completed linmos command that coadded a field

    Returns:
        FieldSummary: The updated field summary object with the linmos fits image added
    """
    assert isinstance(linmos_command, LinmosResult), (
        f"{linmos_command=} is type {type(linmos_command)}, expected LinmosResult"
    )

    image_fits = linmos_command.image_fits
    field_summary = field_summary.with_options(linmos_image=image_fits)

    return field_summary


def update_field_summary(
    field_summary: FieldSummary,
    aegean_outputs: AegeanOutputs | None = None,
    mss: list[MS] | None = None,
    linmos_command: LinmosResult | None = None,
    **kwargs,
) -> FieldSummary:
    """Update an existing `FieldSummary` instance with additional information.

    If special steps are required to be carried out based on a known input they will be.
    Otherwise all additional keyword arguments are passed through to the `FieldSummary.with_options`.

    Args:
        field_summary (FieldSummary): Field summary object to update
        aegean_outputs (Optional[AegeanOutputs], optional): Will add RMS and aegean related properties. Defaults to None.
        mss (Optional[Collection[MS]], optionals): Set of measurement sets to describe
        linmos_command (Optional[LinmosResult], optional): The linmos command created when co-adding all beam images together

    Returns:
        FieldSummary: An updated field summary objects
    """

    if aegean_outputs:
        field_summary = add_rms_information(
            field_summary=field_summary, aegean_outputs=aegean_outputs
        )

    if mss:
        field_summary = add_ms_summaries(field_summary=field_summary, mss=mss)

    if linmos_command:
        field_summary = add_linmos_fits_image(
            field_summary=field_summary, linmos_command=linmos_command
        )

    field_summary = field_summary.with_options(**kwargs)

    logger.info(f"Updated {field_summary=}")
    return field_summary


def create_field_summary(
    mss: list[MS | Path],
    cal_sbid_path: Path | None = None,
    holography_path: Path | None = None,
    aegean_outputs: AegeanOutputs | None = None,
    **kwargs,
) -> FieldSummary:
    """Create a field summary object using a measurement set.

    All other keyword arguments are passed directly through to `FieldSummary`

    Args:
        ms (Union[MS, Path]): Measurement set information will be pulled from
        cal_sbid_path (Optional[Path], optional): Path to an example of a bandpass measurement set. Defaults to None.
        holography_path (Optional[Path], optional): The holography fits cube used (or will be) to linmos. Defaults to None.
        aegean_outputs (Optional[AegeanOutputs], optional): Should RMS / source information be added to the instance. Defaults to None.
        mss (Optional[Collection[MS]], optionals): Set of measurement sets to describe

    Returns:
        FieldSummary: A summary of a field
    """

    logger.info("Creating field summary object")

    mss = [MS.cast(ms=ms) for ms in mss]

    # TODO: A check here to ensure all MSs are in a consistent format
    # and are from the same field
    ms = MS.cast(ms=mss[0])

    ms_components = processed_ms_format(in_name=ms.path)
    assert ms_components is not None, (
        f"MS ({ms.path}) is not in a flint-processed format!"
    )
    sbid = str(ms_components.sbid)
    field = ms_components.field

    assert field is not None, f"Field name is empty in {ms_components=} from {ms.path=}"

    try:
        cal_sbid = (
            str(get_sbid_from_path(path=cal_sbid_path)) if cal_sbid_path else None
        )
    except ValueError:
        cal_sbid = "-9999"
        logger.info(f"Extracting SBID from {cal_sbid_path=} failed. Using {cal_sbid=}")

    ms_times = get_times_from_ms(ms=ms)
    integration = ms_times.ptp().to(u.second).value
    location = get_telescope_location_from_ms(ms=ms)

    pol_axis = _get_pol_axis_as_rad(ms=ms)

    field_summary = FieldSummary(
        sbid=sbid,
        field_name=field,
        cal_sbid=f"{cal_sbid}",  # could be None, and that's OK
        location=location,
        integration_time=integration,
        holography_path=holography_path,
        ms_times=Time([ms_times.min(), ms_times.max()]),
        pol_axis=pol_axis,
        **kwargs,
    )

    field_summary = add_ms_summaries(
        field_summary=field_summary, mss=[MS.cast(ms=ms) for ms in mss]
    )

    if aegean_outputs:
        field_summary = add_rms_information(
            field_summary=field_summary, aegean_outputs=aegean_outputs
        )

    return field_summary


class BeamSummary(NamedTuple):
    """Structure intended to collect components from a measurement set
    as it is being processed through a continuum imaging pipeline
    """

    ms_summary: MSSummary
    """A summary object of a measurement set"""
    image_set: ImageSet | None = None
    """A set of images that have been created from the measurement set represented by `summary`"""
    components: AegeanOutputs | None = None
    """The source finding components from the aegean source finder"""

    def with_options(self, **kwargs) -> BeamSummary:
        prop = self._asdict()
        prop.update(**kwargs)

        return BeamSummary(**prop)


def create_beam_summary(
    ms: MS | Path,
    image_set: ImageSet | WSCleanResult | None = None,
    components: AegeanOutputs | None = None,
) -> BeamSummary:
    """Create a summary of a beam

    Args:
        ms (Union[MS, Path]): The measurement set being considered
        image_set (Optional[ImageSet], optional): Images produced from an imager. Defaults to None.
        components (Optional[AegeanOutputs], optional): Source finding output components. Defaults to None.

    Returns:
        BeamSummary: Summary object of a beam
    """
    ms = MS.cast(ms=ms)
    logger.info(f"Creating BeamSummary for {ms.path=}")
    ms_summary = describe_ms(ms=ms)

    # TODO: Another example where a .cast type method could be useful
    # or where a standardised set of attributes with a HasImageSet type
    if image_set:
        image_set = (
            image_set if isinstance(image_set, ImageSet) else image_set.image_set
        )

    beam_summary = BeamSummary(
        ms_summary=ms_summary, image_set=image_set, components=components
    )

    return beam_summary



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information
from __future__ import annotations

import shutil
from pathlib import Path

import flint

project = "flint"
copyright = "2025, Tim Galvin"
author = "Tim Galvin"
version = release = flint.__version__

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration


extensions = [
    "myst_parser",
    "sphinx.ext.autodoc",
    "sphinx_autodoc_typehints",
    "sphinx.ext.doctest",
    "sphinx.ext.intersphinx",
    "sphinx.ext.todo",
    "sphinx.ext.coverage",
    "sphinx.ext.mathjax",
    "sphinx.ext.ifconfig",
    "sphinx.ext.viewcode",
    "sphinx.ext.githubpages",
    "sphinx.ext.napoleon",
    "sphinx_copybutton",
    "autoapi.extension",
    "sphinxarg.ext",
]

myst_enable_extensions = [
    "colon_fence",
]

autoapi_type = "python"
autoapi_dirs = ["../flint"]
autoapi_member_order = "groupwise"
autoapi_keep_files = False
autoapi_root = "autoapi"
autoapi_add_toctree_entry = True

# Napoleon settings
napoleon_google_docstring = True
napoleon_numpy_docstring = True
napoleon_include_init_with_doc = True
napoleon_include_private_with_doc = True
napoleon_include_special_with_doc = True
napoleon_use_admonition_for_examples = False
napoleon_use_admonition_for_notes = False
napoleon_use_admonition_for_references = False
napoleon_use_ivar = False
napoleon_use_param = True
napoleon_use_rtype = True

source_suffix = [".rst", ".md"]
templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "furo"
# html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]

# copy logo to _static
static_path = Path("_static")
logo_path = Path("../_static/logo.jpeg")

shutil.copyfile(logo_path, static_path / logo_path.name)



"""Some basic checks around the potato peel functionality"""

from __future__ import annotations

import shutil
from pathlib import Path

import astropy.units as u
import numpy as np
import pytest
from astropy.coordinates import SkyCoord
from astropy.table import Table

from flint.imager.wsclean import WSCleanOptions
from flint.options import MS
from flint.peel.potato import (
    NormalisedSources,
    PotatoConfigCommand,
    PotatoConfigOptions,
    PotatoPeelArguments,
    PotatoPeelCommand,
    PotatoPeelOptions,
    _potato_config_command,
    _potato_peel_command,
    find_sources_to_peel,
    get_source_props_from_table,
    load_known_peel_sources,
    source_within_image_fov,
)
from flint.sky_model import (
    AiryResponse,
    GaussianResponse,
    SincSquaredResponse,
    generate_pb,
)
from flint.utils import get_packaged_resource_path


@pytest.fixture
def ms_example(tmpdir):
    ms_zip = Path(
        get_packaged_resource_path(
            package="flint.data.tests",
            filename="SB39400.RACS_0635-31.beam0.small.ms.zip",
        )
    )
    outpath = Path(tmpdir) / "39400"

    shutil.unpack_archive(ms_zip, outpath)

    ms_path = Path(outpath) / "SB39400.RACS_0635-31.beam0.small.ms"

    return ms_path


# TODO: NEED TESTS FOR THE POTATO PEEL OPTIONS
# TODO: NEED TESTS FOR THE POTATO PEEL COMMAND


def test_source_in_image_fov():
    """Test to see if source is within an image FoV"""
    # This comes from a test wsclean produced image. Not sure
    # why or how a minus crval1 has come out.
    wcs_dict = dict(
        NAXIS1=8128,
        NAXIS2=8128,
        ORIGIN="WSClean",
        CTYPE1="RA---SIN",
        CRPIX1=4065,
        CRVAL1=-1.722664244157e02,
        CDELT1=-6.944444444444e-04,
        CUNIT1="deg",
        CTYPE2="DEC--SIN",
        CRPIX2=4065,
        CRVAL2=2.625771981318e00,
        CDELT2=6.944444444444e-04,
        CUNIT2="deg",
    )
    center_position = SkyCoord(wcs_dict["CRVAL1"] * u.deg, wcs_dict["CRVAL2"] * u.deg)
    known_tato = SkyCoord("12:29:06 02:03:08", unit=(u.hourangle, u.deg))
    outside_fov = SkyCoord("12:29:06 10:03:08", unit=(u.hourangle, u.deg))

    in_image = source_within_image_fov(
        source_coord=known_tato,
        beam_coord=center_position,
        image_size=wcs_dict["NAXIS1"],
        pixel_scale=wcs_dict["CRVAL1"] * u.deg,
    )

    assert in_image == True  # noqa

    outside_strs = (
        ("12:29:06 10:03:08", False),
        ("12:29:06 8:03:08", False),
        ("12:29:06 6:03:08", False),
        ("12:29:06 5:30:08", False),
        ("12:29:06 5:27:08", False),
        ("12:29:06 5:26:08", True),
    )

    for outside_str, outside_value in outside_strs:
        outside_fov = SkyCoord(outside_str, unit=(u.hourangle, u.deg))
        out_image = source_within_image_fov(
            source_coord=outside_fov,
            beam_coord=center_position,
            image_size=wcs_dict["NAXIS1"],
            pixel_scale=wcs_dict["CDELT1"] * u.deg,
        )

        assert out_image == outside_value


def test_potato_peel_command(ms_example):
    """Test to see if the potato peel command can be generated correctly"""
    ms = MS(path=ms_example, column="DATA")

    # Set up the sources to peel out
    image_options = WSCleanOptions(size=8000, scale="2.5arcsec")
    sources = find_sources_to_peel(ms=ms, image_options=image_options)
    source_props = get_source_props_from_table(table=sources)

    potato_peel_arguments = PotatoPeelArguments(
        ms=ms.path,
        ras=source_props.source_ras,
        decs=source_props.source_decs,
        peel_fovs=source_props.source_fovs,
        n=source_props.source_names,
        image_fov=1.0,
    )
    potato_peel_options = PotatoPeelOptions()

    peel_command = _potato_peel_command(
        ms=ms,
        potato_peel_arguments=potato_peel_arguments,
        potato_peel_options=potato_peel_options,
    )

    assert isinstance(peel_command, PotatoPeelCommand)
    assert peel_command.ms.path == ms.path
    assert isinstance(peel_command.command, str)

    # expected_command = f"hot_potato {str(ms.path)} 1.0000 --ras 83.82499999999999 79.94999999999999 --decs -5.386388888888889 -45.764722222222225 --peel_fovs 0.11850000000000001 0.105 -n Orion_A Pictor_A -solint 30 -calmode P -minpeelflux 0.5 -refant 1 --direct_subtract --intermediate_peels -T peel "
    # assert peel_command.command == expected_command


def test_potato_config_command():
    a = PotatoConfigOptions()
    ex = Path("This/example/SB1234.potato.config")

    command = _potato_config_command(config_path=ex, potato_config_options=a)
    # expected = "peel_configuration.py This/example/SB1234.potato.config --image_size 6148 --image_scale 0.0006944 --image_briggs -1.5 --image_channels 4 --image_minuvl 0.0 --peel_size 1000 --peel_scale 0.0006944 --peel_channels 16 --peel_nmiter 7 --peel_minuvl 0.0 --peel_multiscale "

    assert isinstance(command, PotatoConfigCommand)
    # assert command.command == expected
    assert command.config_path == ex


def test_normalised_sources_to_peel(ms_example):
    """See whether the normalisation of sources in a table for potato CLI works"""
    image_options = WSCleanOptions(size=8000, scale="2.5arcsec")

    sources = find_sources_to_peel(ms=ms_example, image_options=image_options)

    source_props = get_source_props_from_table(table=sources)

    assert isinstance(source_props, NormalisedSources)
    assert isinstance(source_props.source_ras, tuple)
    assert isinstance(source_props.source_decs, tuple)
    assert isinstance(source_props.source_fovs, tuple)
    assert isinstance(source_props.source_names, tuple)

    assert len(source_props.source_ras) == 3
    assert len(source_props.source_decs) == 3
    assert len(source_props.source_fovs) == 3
    assert len(source_props.source_names) == 3


def test_check_sources_to_peel(ms_example):
    """See whether the peeling constraints work"""
    image_options = WSCleanOptions(size=8128, scale="2.5arcsec")

    sources = find_sources_to_peel(ms=ms_example, image_options=image_options)

    known_sources = load_known_peel_sources()

    assert len(known_sources) > len(sources)

    # See the WCS constructed in the above
    center_position = SkyCoord(-1.722664244157e02 * u.deg, 2.625771981318e00 * u.deg)
    sources = find_sources_to_peel(
        ms=ms_example,
        image_options=image_options,
        override_beam_position_with=center_position,
    )

    assert len(sources) == 3
    assert sources["Name"][0] == "VirA"


def test_load_peel_sources():
    """Ensure we can load in the reference set of sources for peeling"""

    tab = load_known_peel_sources()
    assert isinstance(tab, Table)
    assert len(tab) > 4


def test_pb_models():
    """See if the responses can be loaded and executed"""
    freq = 1.0 * u.GHz
    aperture = 12 * u.m
    offset = np.arange(0, 200) * u.arcmin

    pb_gauss = generate_pb(
        pb_type="gaussian", freqs=freq, aperture=aperture, offset=offset
    )
    pb_sinc = generate_pb(
        pb_type="sincsquared", freqs=freq, aperture=aperture, offset=offset
    )
    pb_airy = generate_pb(pb_type="airy", freqs=freq, aperture=aperture, offset=offset)

    assert isinstance(pb_gauss, GaussianResponse)
    assert isinstance(pb_sinc, SincSquaredResponse)
    assert isinstance(pb_airy, AiryResponse)

    assert np.isclose(pb_gauss.atten[0], 1.0)
    assert np.isclose(pb_sinc.atten[0], 1.0)
    assert np.isclose(pb_airy.atten[0], 1.0)

    for m in (pb_gauss, pb_sinc, pb_airy):
        assert m.freqs == freq

    # Following known values were generated with the following
    # code, verified, and placed here
    """
    freq = 1.0 * u.GHz
    aperture = 12 * u.m
    offset = np.arange(0, 200) * u.arcmin

    pb_gauss = generate_pb(
        pb_type="gaussian", freqs=freq, aperture=aperture, offset=offset
    )
    pb_sinc = generate_pb(
        pb_type="sincsquared", freqs=freq, aperture=aperture, offset=offset
    )
    pb_airy = generate_pb(pb_type="airy", freqs=freq, aperture=aperture, offset=offset)
    """

    assert np.allclose(
        pb_gauss.atten[10:15],
        [0.96310886, 0.95553634, 0.94731093, 0.93845055, 0.9289744],
    )
    assert np.allclose(
        pb_sinc.atten[10:15],
        [0.96516646, 0.95797617, 0.95015023, 0.94170175, 0.93264483],
    )
    assert np.allclose(
        pb_airy.atten[10:15],
        [0.96701134, 0.96020018, 0.95278619, 0.94478158, 0.93619952],
    )

    assert np.allclose(
        pb_gauss.atten[140:145],
        [0.0006315, 0.0005682, 0.00051086, 0.00045896, 0.00041203],
    )
    assert np.allclose(
        pb_sinc.atten[140:145],
        [0.04699694, 0.04675459, 0.04641816, 0.04599033, 0.04547408],
    )
    assert np.allclose(
        pb_airy.atten[140:145],
        [0.01749399, 0.01748965, 0.01743916, 0.01734373, 0.01720478],
    )


def test_bad_pb_model():
    """Makes sure ValueError raised on unknown model"""
    freq = 1.0 * u.GHz
    aperture = 12 * u.m
    offset = np.arange(0, 200) * u.arcmin

    with pytest.raises(ValueError):
        generate_pb(
            pb_type="ThisDoesNotExist", freqs=freq, aperture=aperture, offset=offset
        )



"""Basic tests around prefect helper functions"""

from __future__ import annotations

from prefect import flow
from prefect.logging import disable_run_logger
from prefect.testing.utilities import prefect_test_harness

from flint.prefect.helpers import enable_loguru_support


@flow
def example_flow():
    enable_loguru_support()
    return "JackSparrow"


def test_enable_loguru_support():
    """Some packages may be using loguru (e.g. crystalball). Should
    we want those logs to be captured we need to modify the loguru
    logger. A helpful function has been added to this end. This
    is a simple, very basic test to make sure it can still run without
    error, though whether it still works is a completely different
    question!"""

    with prefect_test_harness(), disable_run_logger():
        assert example_flow() == "JackSparrow"



"""Gemeral help functions that could be used for self-calibration
across different packages.
"""

from __future__ import annotations

from math import ceil
from pathlib import Path

from flint.logging import logger
from flint.ms import get_freqs_from_ms
from flint.options import MS
from flint.utils import remove_files_folders


def create_and_check_caltable_path(
    ms: MS, channel_range: tuple[int, int] | None = None, remove_if_exists: bool = False
) -> Path:
    """Create the output name path for the gaincal solutions table.

    If the table already exists it will be removed.

    Args:
        cal_ms (MS): A description of the measurement set
        channel_range (tuple[int,int] | None, optional): Channel start and end, which will be appended. Defaults to None.
        remove_if_exists (bool, optional): If ``True`` and the table exists, remove it. Defaults to False.

    Returns:
        Path: Output path of the solutions table
    """

    cal_suffix = ".caltable"
    if channel_range:
        cal_suffix += f".ch{channel_range[0]:04d}-{channel_range[1]:04d}"
    cal_table_name = ms.path.with_suffix(cal_suffix)

    cal_table = ms.path.absolute().parent / cal_table_name
    logger.info(f"Will create calibration table {cal_table}.")

    if remove_if_exists and cal_table.exists():
        logger.warning(f"Removing {cal_table!s}")
        remove_files_folders(cal_table)

    return cal_table


def get_channel_ranges_given_nspws(
    num_channels: int, nspws: int
) -> tuple[tuple[int, int], ...]:
    """Given the number of channels construct the start and end channel indices
    for the specified number of spectral windows. The interval step size will be ceiled
    should the the ``num_channels / nspw`` not be whole. In this case the last
    interval will be smaller than the others.

    Args:
        num_channels (int): The number of channels spanning some band
        nspws (int): The number of segments across the band

    Returns:
        tuple[tuple[int,int]]: The start and end channel index spanning the number of channels.
    """

    step = ceil(num_channels / nspws)
    starts = list(range(0, num_channels, step))
    ends = [min(start + step - 1, num_channels - 1) for start in starts]

    return tuple(zip(starts, ends))


def get_channel_ranges_given_nspws_for_ms(
    ms: MS | Path, nspw: int
) -> tuple[tuple[int, int], ...]:
    """Construct channel range intervals for the channels in a measurement set
    given a desired number of spectral windows

    Args:
        ms (MS | Path): The measurement set to construct channel ranges for
        nspw (int): Number of channel intervals to construct

    Returns:
        tuple[tuple[int,int], ...]: The collection of start and end channel intervals. The length will be ``nspw``
    """
    ms = MS.cast(ms=ms)

    logger.info(f"Considering {ms.path}, obtaining channel ranges for {nspw=}")
    freqs = get_freqs_from_ms(ms=ms)

    return get_channel_ranges_given_nspws(num_channels=len(freqs), nspws=nspw)


def consider_skip_selfcal_on_round(
    current_round: int, skip_selfcal_on_rounds: int | list[int] | None
) -> bool:
    """Consider whether the self-calibration process (derive and applying solutions)
    should be skipped on a particular imaging round.

    Should `current_round` be in `skip_selfcal_on_round` then the self-calibration
    should be skipped and a `True` is returned.

    Args:
        current_round (int): The current imaging round being considered
        skip_selfcal_on_rounds (Union[int, List[int], None]): The set of rounds that should be considerede for skipping. If None a False is returned.

    Returns:
        bool: Whether the round is skipped (True) or performed (False)
    """

    # Consider whether this is unset, in which case all rounds should be self-cal
    if skip_selfcal_on_rounds is None:
        return False

    # For sanity consider a single int case
    skip_selfcal_on_rounds = (
        skip_selfcal_on_rounds
        if isinstance(skip_selfcal_on_rounds, list)
        else [skip_selfcal_on_rounds]
    )

    return current_round in skip_selfcal_on_rounds
